# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
import abc
from functools import partial
from typing import Any, Callable, Dict, List, Optional, Tuple

import torch
import torchmetrics
from pytorch_lightning.loggers.logger import Logger
from pytorch_lightning.utilities.types import STEP_OUTPUT
from torch.nn import Parameter
from torch.optim import Optimizer
from torch.utils.data import DataLoader, Dataset, Subset

from renate import defaults
from renate.data.datasets import _EnumeratedDataset, _TransformedDataset
from renate.memory.buffer import DataDict
from renate.models import RenateModule
from renate.types import NestedTensors
from renate.updaters.learner import ReplayLearner
from renate.updaters.learner_components.component import Component
from renate.updaters.learner_components.losses import (
    WeightedCLSLossComponent,
    WeightedCustomLossComponent,
    WeightedMeanSquaredErrorLossComponent,
    WeightedPooledOutputDistillationLossComponent,
)
from renate.updaters.learner_components.reinitialization import (
    ShrinkAndPerturbReinitializationComponent,
)
from renate.updaters.model_updater import SingleTrainingLoopUpdater
from renate.utils.misc import maybe_populate_mask_and_ignore_logits
from renate.utils.pytorch import move_tensors_to_device


class BaseExperienceReplayLearner(ReplayLearner, abc.ABC):
    """A base implementation of experience replay.

    It is designed for the online CL setting, where only one pass over each new chunk of data is
    allowed. The Learner maintains a Reservoir buffer. In the training step, it samples a batch of
    data from the memory and appends it to the batch of current-task data. At the end of the
    training step, the memory is updated.

    Args:
        components: An ordered dictionary of components that are part of the experience replay
            learner.
        loss_weight: A scalar weight factor for the base loss function to trade it off with other
            loss functions added by `components`.
        ema_memory_update_gamma: The gamma used for exponential moving average to update the meta
            data with respect to the logits and intermediate representation, if there is some.
        loss_normalization: Whether to normalize the loss by the weights of all the components.
    """

    def __init__(
        self,
        components: Dict[str, Component],
        loss_weight: float = defaults.LOSS_WEIGHT,
        ema_memory_update_gamma: float = defaults.EMA_MEMORY_UPDATE_GAMMA,
        loss_normalization: int = defaults.LOSS_NORMALIZATION,
        **kwargs: Any,
    ) -> None:
        self._components_names = list(components.keys())
        super().__init__(**kwargs)
        self._memory_loader: Optional[DataLoader] = None
        self._components = components
        self._loss_weight = loss_weight
        self._ema_memory_update_gamma = ema_memory_update_gamma
        self._use_loss_normalization = bool(loss_normalization)

    def _create_metrics_collections(
        self, logged_metrics: Optional[Dict[str, torchmetrics.Metric]] = None
    ) -> None:
        super()._create_metrics_collections(logged_metrics)
        for name in self._components_names:
            if name in self._loss_collections:
                raise ValueError(
                    f"Component name {name} is already used as a loss name. Please pick a "
                    "different name."
                )
            self._loss_collections["train_losses"].update({name: torchmetrics.MeanMetric()})

    def on_model_update_start(
        self,
        train_dataset: Dataset,
        val_dataset: Dataset,
        train_dataset_collate_fn: Optional[Callable] = None,
        val_dataset_collate_fn: Optional[Callable] = None,
        task_id: Optional[str] = None,
    ) -> None:
        """Called before a model update starts."""
        super().on_model_update_start(
            train_dataset=train_dataset,
            val_dataset=val_dataset,
            train_dataset_collate_fn=train_dataset_collate_fn,
            val_dataset_collate_fn=val_dataset_collate_fn,
            task_id=task_id,
        )
        self._set_memory_loader()

    def train_dataloader(self) -> DataLoader:
        train_dataset = _EnumeratedDataset(
            _TransformedDataset(
                self._train_dataset,
                transform=self._train_transform,
                target_transform=self._train_target_transform,
            )
        )
        return DataLoader(
            train_dataset,
            batch_size=self._batch_size,
            shuffle=True,
            generator=self._rng,
            pin_memory=True,
            collate_fn=self._train_collate_fn,
        )

    def on_train_start(self) -> None:
        """PyTorch Lightning function to be run at the start of the training."""
        super().on_train_start()
        for component in self._components.values():
            component.on_train_start(model=self._model)

    def training_step(
        self, batch: Tuple[torch.Tensor, Tuple[NestedTensors, torch.Tensor]], batch_idx: int
    ) -> STEP_OUTPUT:
        """PyTorch Lightning function to return the training loss."""
        idx, (inputs, targets) = batch
        step_output = super().training_step(batch=(inputs, targets), batch_idx=batch_idx)
        step_output["train_data_idx"] = idx
        step_output["loss"] *= self._loss_weight

        batch_memory: Optional[torch.Tensor] = None
        metadata_memory: Optional[torch.Tensor] = None
        outputs_memory: Optional[torch.Tensor] = None
        intermediate_representation_memory: Optional[List[torch.Tensor]] = None
        loss_normalization = self._loss_weight
        if self._memory_loader is not None:
            for name, component in self._components.items():
                memory_sampled = False
                if component.sample_new_memory_batch or batch_memory is None:
                    batch_memory = self._sample_from_buffer(device=step_output["loss"].device)
                    (inputs_memory, _), metadata_memory = batch_memory
                    outputs_memory = self(inputs_memory)

                    outputs_memory, self._class_mask = maybe_populate_mask_and_ignore_logits(
                        self._mask_unused_classes,
                        self._class_mask,
                        self._classes_in_current_task,
                        outputs_memory,
                    )
                    intermediate_representation_memory = (
                        self._model.get_intermediate_representation()
                    )
                    self._model.reset_intermediate_representation_cache()
                    memory_sampled = True

                component_loss = component.loss(
                    outputs_memory=outputs_memory,
                    batch_memory=batch_memory,
                    intermediate_representation_memory=intermediate_representation_memory,
                ).mean()
                self._loss_collections["train_losses"][name](component_loss)
                step_output["loss"] += component_loss
                loss_normalization += component.weight
                if memory_sampled and self._ema_memory_update_gamma < 1.0:
                    mem_idx = metadata_memory["idx"].cpu()
                    self._memory_buffer.metadata["outputs"][mem_idx] = self._memory_buffer.metadata[
                        "outputs"
                    ][
                        mem_idx
                    ].cpu() * self._ema_memory_update_gamma + outputs_memory.detach().cpu() * (
                        1.0 - self._ema_memory_update_gamma
                    )
            if self._use_loss_normalization:
                step_output["loss"] /= loss_normalization

        return step_output

    def _sample_from_buffer(self, device: torch.device) -> Optional[Tuple[NestedTensors, DataDict]]:
        """Function to sample from the buffer, if buffer is populated."""
        if self._memory_loader is not None and len(self._memory_buffer) >= self._memory_batch_size:
            memory_batch = next(iter(self._memory_loader))
            return move_tensors_to_device(memory_batch, device)
        else:
            return None

    def training_step_end(self, step_output: STEP_OUTPUT) -> STEP_OUTPUT:
        """PyTorch Lightning function to perform after the training step."""
        super().training_step_end(step_output)
        self._update_memory_buffer(step_output)
        return step_output

    def _update_memory_buffer(self, step_output: STEP_OUTPUT) -> None:
        outputs = step_output["outputs"]
        metadata = {"outputs": outputs.detach().cpu()}
        for i, intermediate_representation in enumerate(step_output["intermediate_representation"]):
            metadata[
                f"intermediate_representation_{i}"
            ] = intermediate_representation.detach().cpu()
        # Some datasets have problems using tensors as subset indices, convert to list of ints.
        train_data_idx = [int(idx) for idx in step_output["train_data_idx"]]
        dataset = Subset(self._train_dataset, train_data_idx)
        self._memory_buffer.update(dataset, metadata)
        self._set_memory_loader()

    def _set_memory_loader(self) -> None:
        """Create a memory loader from a memory buffer."""
        if self._memory_loader is None and len(self._memory_buffer) >= self._memory_batch_size:
            self._memory_loader = DataLoader(
                dataset=self._memory_buffer,
                batch_size=self._memory_batch_size,
                drop_last=True,
                shuffle=True,
                generator=self._rng,
                pin_memory=True,
                collate_fn=self._train_collate_fn,
            )

    def on_train_batch_end(self, outputs: STEP_OUTPUT, batch: Any, batch_idx: int) -> None:
        """PyTorch Lightning function to perform after the training and optimizer step."""
        super().on_train_batch_end(outputs, batch, batch_idx)
        for component in self._components.values():
            component.on_train_batch_end(model=self._model)

    @abc.abstractmethod
    def components(self, **kwargs) -> Dict[str, Component]:
        """Returns the components of the learner.

        This is a user-defined function that should return a dictionary of components.
        """

    def on_load_checkpoint(self, checkpoint: Dict[str, Any]) -> None:
        """Load states of components."""
        super().on_load_checkpoint(checkpoint)
        for component in self._components.values():
            component.on_load_checkpoint(checkpoint)

    def on_save_checkpoint(self, checkpoint: Dict[str, Any]) -> None:
        """Save states of components."""
        super().on_save_checkpoint(checkpoint)
        for component in self._components.values():
            component.on_save_checkpoint(checkpoint)


class ExperienceReplayLearner(BaseExperienceReplayLearner):
    """This is the version of experience replay proposed in

    Chaudhry, Arslan, et al. "On tiny episodic memories in continual learning."
    arXiv preprint arXiv:1902.10486 (2019).

    Args:
        alpha: The weight of the cross-entropy loss component applied to the memory samples.
    """

    def __init__(self, alpha: float = defaults.ER_ALPHA, **kwargs) -> None:
        components = self.components(loss_fn=kwargs["loss_fn"], alpha=alpha)
        super().__init__(components=components, **kwargs)

    def components(
        self, loss_fn: Optional[torch.nn.Module] = None, alpha: float = defaults.ER_ALPHA
    ) -> Dict[str, Component]:
        return {
            "memory_loss": WeightedCustomLossComponent(
                loss_fn=loss_fn, weight=alpha, sample_new_memory_batch=True
            )
        }


class DarkExperienceReplayLearner(ExperienceReplayLearner):
    """A Learner that implements Dark Experience Replay.

    Pietro Buzzega, Matteo Boschini, Angelo Porrello, Davide Abati, Simone Calderara:
    Dark Experience for General Continual Learning: a Strong, Simple Baseline. NeurIPS 2020

    Args:
        alpha: The weight of the mean squared error loss component between memorised logits and the
            current logits on the memory data.
        beta: The weight of the cross-entropy loss component between memorised targets and the
            current logits on the memory data.
    """

    def __init__(
        self, alpha: float = defaults.DER_ALPHA, beta: float = defaults.DER_BETA, **kwargs
    ) -> None:
        super().__init__(alpha=beta, **kwargs)
        self._components = self.components(loss_fn=kwargs["loss_fn"], alpha=alpha, beta=beta)

    def components(
        self,
        loss_fn: Optional[torch.nn.Module] = None,
        alpha: float = defaults.DER_ALPHA,
        beta: float = defaults.DER_BETA,
    ) -> Dict[str, Component]:
        components = super().components(loss_fn=loss_fn, alpha=beta)
        components.update(
            {
                "mse_loss": WeightedMeanSquaredErrorLossComponent(
                    weight=alpha, sample_new_memory_batch=False
                )
            }
        )
        return components


class PooledOutputDistillationExperienceReplayLearner(BaseExperienceReplayLearner):
    """A Learner that implements Pooled Output Distillation.

    Douillard, Arthur, et al. "Podnet: Pooled outputs distillation for small-tasks incremental
    learning."
    European Conference on Computer Vision. Springer, Cham, 2020.

    Args:
        alpha: Scaling value which scales the loss with respect to all intermediate representations.
        distillation_type: Which distillation type to apply with respect to the intermediate
            representation.
        normalize: Whether to normalize both the current and cached features before computing the
            Frobenius norm.
    """

    def __init__(
        self,
        alpha: float = defaults.POD_ALPHA,
        distillation_type: str = defaults.POD_DISTILLATION_TYPE,
        normalize: bool = defaults.POD_NORMALIZE,
        **kwargs,
    ) -> None:
        components = self.components(
            alpha=alpha, distillation_type=distillation_type, normalize=normalize
        )
        super().__init__(components=components, **kwargs)

    def components(
        self,
        alpha: float = defaults.POD_ALPHA,
        distillation_type: str = defaults.POD_DISTILLATION_TYPE,
        normalize: bool = defaults.POD_NORMALIZE,
    ) -> Dict[str, Component]:
        return {
            "pod_loss": WeightedPooledOutputDistillationLossComponent(
                weight=alpha,
                sample_new_memory_batch=True,
                distillation_type=distillation_type,
                normalize=normalize,
            )
        }


class CLSExperienceReplayLearner(BaseExperienceReplayLearner):
    """A learner that implements a Complementary Learning Systems Based Experience Replay.

    Arani, Elahe, Fahad Sarfraz, and Bahram Zonooz.
    "Learning fast, learning slow: A general continual learning method based on complementary
    learning system."
    arXiv preprint arXiv:2201.12604 (2022).

    Args:
        alpha: Scaling value for the cross-entropy loss.
        beta: Scaling value for the consistency loss.
        stable_model_update_weight: The starting weight for the exponential moving average to update
            the stable model copy.
        plastic_model_update_weight: The starting weight for the exponential moving average to
            update the plastic model copy.
        stable_model_update_probability: The probability to update the stable model copy.
        plastic_model_update_probability: The probability to update the plastic model copy.
    """

    def __init__(
        self,
        alpha: float = defaults.CLS_ALPHA,
        beta: float = defaults.CLS_BETA,
        stable_model_update_weight: float = defaults.CLS_STABLE_MODEL_UPDATE_WEIGHT,
        plastic_model_update_weight: float = defaults.CLS_PLASTIC_MODEL_UPDATE_WEIGHT,
        stable_model_update_probability: float = defaults.CLS_STABLE_MODEL_UPDATE_PROBABILITY,
        plastic_model_update_probability: float = defaults.CLS_PLASTIC_MODEL_UPDATE_PROBABILITY,
        **kwargs,
    ):
        components = self.components(
            model=kwargs["model"],
            loss_fn=kwargs["loss_fn"],
            alpha=alpha,
            beta=beta,
            stable_model_update_weight=stable_model_update_weight,
            plastic_model_update_weight=plastic_model_update_weight,
            stable_model_update_probability=stable_model_update_probability,
            plastic_model_update_probability=plastic_model_update_probability,
        )

        super().__init__(components=components, **kwargs)

    def components(
        self,
        model: RenateModule,
        loss_fn: torch.nn.Module,
        alpha: float = defaults.CLS_ALPHA,
        beta: float = defaults.CLS_BETA,
        plastic_model_update_weight: float = defaults.CLS_PLASTIC_MODEL_UPDATE_WEIGHT,
        stable_model_update_weight: float = defaults.CLS_STABLE_MODEL_UPDATE_WEIGHT,
        plastic_model_update_probability: float = defaults.CLS_PLASTIC_MODEL_UPDATE_PROBABILITY,
        stable_model_update_probability: float = defaults.CLS_STABLE_MODEL_UPDATE_PROBABILITY,
    ) -> Dict[str, Component]:
        return {
            "memory_loss": WeightedCustomLossComponent(
                loss_fn=loss_fn, weight=alpha, sample_new_memory_batch=True
            ),
            "cls_loss": WeightedCLSLossComponent(
                weight=beta,
                sample_new_memory_batch=False,
                model=model,
                plastic_model_update_weight=plastic_model_update_weight,
                stable_model_update_weight=stable_model_update_weight,
                plastic_model_update_probability=plastic_model_update_probability,
                stable_model_update_probability=stable_model_update_probability,
            ),
        }


class SuperExperienceReplayLearner(BaseExperienceReplayLearner):
    """A learner that implements a selected combination of methods.

    Args:
        der_alpha: The weight of the mean squared error loss component between memorised logits and
            the current logits on the memory data.
        der_beta: The weight of the cross-entropy loss component between memorised targets and the
            current logits on the memory data.
        sp_shrink_factor: Shrinking value applied with respect to shrink and perturbation.
        sp_sigma: Standard deviation applied with respect to shrink and perturbation.
        cls_alpha: Scaling value for the consistency loss added to the base cross-entropy loss.
        cls_stable_model_update_weight: The starting weight for the exponential moving average to
            update the stable model copy.
        cls_plastic_model_update_weight: The starting weight for the exponential moving average to
            update the plastic model copy.
        cls_stable_model_update_probability: The probability to update the stable model copy.
        cls_plastic_model_update_probability: The probability to update the plastic model copy.
        pod_alpha: Scaling value which scales the loss with respect to all intermediate
            representations.
        pod_distillation_type: Which distillation type to apply with respect to the intermediate
            representation.
        pod_normalize: Whether to normalize both the current and cached features before computing
            the Frobenius norm.
        ema_memory_update_gamma: The gamma used for exponential moving average to update the meta
            data with respect to the logits and intermediate representation, if there is some.
    """

    def __init__(
        self,
        der_alpha: float = defaults.SER_DER_ALPHA,
        der_beta: float = defaults.SER_DER_BETA,
        sp_shrink_factor: float = defaults.SER_SP_SHRINK_FACTOR,
        sp_sigma: float = defaults.SER_SP_SIGMA,
        cls_alpha: float = defaults.SER_CLS_ALPHA,
        cls_stable_model_update_weight: float = defaults.SER_CLS_STABLE_MODEL_UPDATE_WEIGHT,
        cls_plastic_model_update_weight: float = defaults.SER_CLS_PLASTIC_MODEL_UPDATE_WEIGHT,
        cls_stable_model_update_probability: float = defaults.SER_CLS_STABLE_MODEL_UPDATE_PROBABILITY,  # noqa: E501
        cls_plastic_model_update_probability: float = defaults.SER_CLS_PLASTIC_MODEL_UPDATE_PROBABILITY,  # noqa: E501
        pod_alpha: float = defaults.SER_POD_ALPHA,
        pod_distillation_type: str = defaults.SER_POD_DISTILLATION_TYPE,
        pod_normalize: bool = defaults.SER_POD_NORMALIZE,
        ema_memory_update_gamma: float = defaults.EMA_MEMORY_UPDATE_GAMMA,
        **kwargs,
    ) -> None:
        components = self.components(
            model=kwargs["model"],
            loss_fn=kwargs["loss_fn"],
            der_alpha=der_alpha,
            der_beta=der_beta,
            sp_shrink_factor=sp_shrink_factor,
            sp_sigma=sp_sigma,
            cls_alpha=cls_alpha,
            cls_stable_model_update_weight=cls_stable_model_update_weight,
            cls_plastic_model_update_weight=cls_plastic_model_update_weight,
            cls_stable_model_update_probability=cls_stable_model_update_probability,
            cls_plastic_model_update_probability=cls_plastic_model_update_probability,
            pod_alpha=pod_alpha,
            pod_distillation_type=pod_distillation_type,
            pod_normalize=pod_normalize,
        )
        super().__init__(
            components=components,
            ema_memory_update_gamma=ema_memory_update_gamma,
            **kwargs,
        )

    def components(
        self,
        model: RenateModule,
        loss_fn: torch.nn.Module,
        der_alpha: float = defaults.SER_DER_ALPHA,
        der_beta: float = defaults.SER_DER_BETA,
        sp_shrink_factor: float = defaults.SER_SP_SHRINK_FACTOR,
        sp_sigma: float = defaults.SER_SP_SIGMA,
        cls_alpha: float = defaults.SER_CLS_ALPHA,
        cls_stable_model_update_weight: float = defaults.SER_CLS_STABLE_MODEL_UPDATE_WEIGHT,
        cls_plastic_model_update_weight: float = defaults.SER_CLS_PLASTIC_MODEL_UPDATE_WEIGHT,
        cls_stable_model_update_probability: float = defaults.SER_CLS_STABLE_MODEL_UPDATE_PROBABILITY,  # noqa: E501
        cls_plastic_model_update_probability: float = defaults.SER_CLS_PLASTIC_MODEL_UPDATE_PROBABILITY,  # noqa: E501
        pod_alpha: float = defaults.SER_POD_ALPHA,
        pod_distillation_type: str = defaults.SER_POD_DISTILLATION_TYPE,
        pod_normalize: bool = defaults.SER_POD_NORMALIZE,
    ) -> Dict[str, Component]:
        return {
            "mse_loss": WeightedMeanSquaredErrorLossComponent(
                weight=der_alpha, sample_new_memory_batch=True
            ),
            "memory_loss": WeightedCustomLossComponent(
                loss_fn=loss_fn, weight=der_beta, sample_new_memory_batch=True
            ),
            "cls_loss": WeightedCLSLossComponent(
                weight=cls_alpha,
                sample_new_memory_batch=False,
                model=model,
                stable_model_update_weight=cls_stable_model_update_weight,
                plastic_model_update_weight=cls_plastic_model_update_weight,
                stable_model_update_probability=cls_stable_model_update_probability,
                plastic_model_update_probability=cls_plastic_model_update_probability,
            ),
            "shrink_perturb": ShrinkAndPerturbReinitializationComponent(
                shrink_factor=sp_shrink_factor, sigma=sp_sigma
            ),
            "pod_loss": WeightedPooledOutputDistillationLossComponent(
                weight=pod_alpha,
                sample_new_memory_batch=True,
                distillation_type=pod_distillation_type,
                normalize=pod_normalize,
            ),
        }


class ExperienceReplayModelUpdater(SingleTrainingLoopUpdater):
    def __init__(
        self,
        model: RenateModule,
        loss_fn: torch.nn.Module,
        optimizer: Callable[[List[Parameter]], Optimizer],
        memory_size: int,
        batch_memory_frac: int = defaults.BATCH_MEMORY_FRAC,
        loss_weight: float = defaults.LOSS_WEIGHT,
        ema_memory_update_gamma: float = defaults.EMA_MEMORY_UPDATE_GAMMA,
        loss_normalization: int = defaults.LOSS_NORMALIZATION,
        alpha: float = defaults.ER_ALPHA,
        learning_rate_scheduler: Optional[partial] = None,
        learning_rate_scheduler_interval: defaults.SUPPORTED_LR_SCHEDULER_INTERVAL_TYPE = defaults.LR_SCHEDULER_INTERVAL,  # noqa: E501
        batch_size: int = defaults.BATCH_SIZE,
        input_state_folder: Optional[str] = None,
        output_state_folder: Optional[str] = None,
        max_epochs: int = defaults.MAX_EPOCHS,
        train_transform: Optional[Callable] = None,
        train_target_transform: Optional[Callable] = None,
        test_transform: Optional[Callable] = None,
        test_target_transform: Optional[Callable] = None,
        buffer_transform: Optional[Callable] = None,
        buffer_target_transform: Optional[Callable] = None,
        metric: Optional[str] = None,
        mode: defaults.SUPPORTED_TUNING_MODE_TYPE = "min",
        logged_metrics: Optional[Dict[str, torchmetrics.Metric]] = None,
        early_stopping_enabled: bool = False,
        logger: Logger = defaults.LOGGER(**defaults.LOGGER_KWARGS),
        accelerator: defaults.SUPPORTED_ACCELERATORS_TYPE = defaults.ACCELERATOR,
        devices: Optional[int] = None,
        strategy: str = defaults.DISTRIBUTED_STRATEGY,
        precision: str = defaults.PRECISION,
        seed: int = defaults.SEED,
        deterministic_trainer: bool = defaults.DETERMINISTIC_TRAINER,
        gradient_clip_val: Optional[float] = defaults.GRADIENT_CLIP_VAL,
        gradient_clip_algorithm: Optional[str] = defaults.GRADIENT_CLIP_ALGORITHM,
        mask_unused_classes: bool = defaults.MASK_UNUSED_CLASSES,
    ):
        learner_kwargs = {
            "memory_size": memory_size,
            "batch_memory_frac": batch_memory_frac,
            "loss_weight": loss_weight,
            "ema_memory_update_gamma": ema_memory_update_gamma,
            "loss_normalization": loss_normalization,
            "alpha": alpha,
            "batch_size": batch_size,
            "seed": seed,
        }
        super().__init__(
            model,
            loss_fn=loss_fn,
            optimizer=optimizer,
            learner_class=ExperienceReplayLearner,
            learner_kwargs=learner_kwargs,
            input_state_folder=input_state_folder,
            output_state_folder=output_state_folder,
            max_epochs=max_epochs,
            learning_rate_scheduler=learning_rate_scheduler,
            learning_rate_scheduler_interval=learning_rate_scheduler_interval,
            train_transform=train_transform,
            train_target_transform=train_target_transform,
            test_transform=test_transform,
            test_target_transform=test_target_transform,
            buffer_transform=buffer_transform,
            buffer_target_transform=buffer_target_transform,
            metric=metric,
            mode=mode,
            logged_metrics=logged_metrics,
            early_stopping_enabled=early_stopping_enabled,
            logger=logger,
            accelerator=accelerator,
            devices=devices,
            strategy=strategy,
            precision=precision,
            deterministic_trainer=deterministic_trainer,
            gradient_clip_algorithm=gradient_clip_algorithm,
            gradient_clip_val=gradient_clip_val,
            mask_unused_classes=mask_unused_classes,
        )


class DarkExperienceReplayModelUpdater(SingleTrainingLoopUpdater):
    def __init__(
        self,
        model: RenateModule,
        loss_fn: torch.nn.Module,
        optimizer: Callable[[List[Parameter]], Optimizer],
        memory_size: int,
        batch_memory_frac: int = defaults.BATCH_MEMORY_FRAC,
        loss_weight: float = defaults.LOSS_WEIGHT,
        ema_memory_update_gamma: float = defaults.EMA_MEMORY_UPDATE_GAMMA,
        loss_normalization: int = defaults.LOSS_NORMALIZATION,
        alpha: float = defaults.DER_ALPHA,
        beta: float = defaults.DER_BETA,
        learning_rate_scheduler: Optional[partial] = None,
        learning_rate_scheduler_interval: defaults.SUPPORTED_LR_SCHEDULER_INTERVAL_TYPE = defaults.LR_SCHEDULER_INTERVAL,  # noqa: E501
        batch_size: int = defaults.BATCH_SIZE,
        input_state_folder: Optional[str] = None,
        output_state_folder: Optional[str] = None,
        max_epochs: int = defaults.MAX_EPOCHS,
        train_transform: Optional[Callable] = None,
        train_target_transform: Optional[Callable] = None,
        test_transform: Optional[Callable] = None,
        test_target_transform: Optional[Callable] = None,
        buffer_transform: Optional[Callable] = None,
        buffer_target_transform: Optional[Callable] = None,
        metric: Optional[str] = None,
        mode: defaults.SUPPORTED_TUNING_MODE_TYPE = "min",
        logged_metrics: Optional[Dict[str, torchmetrics.Metric]] = None,
        early_stopping_enabled: bool = False,
        logger: Logger = defaults.LOGGER(**defaults.LOGGER_KWARGS),
        accelerator: defaults.SUPPORTED_ACCELERATORS_TYPE = defaults.ACCELERATOR,
        devices: Optional[int] = None,
        strategy: str = defaults.DISTRIBUTED_STRATEGY,
        precision: str = defaults.PRECISION,
        seed: int = defaults.SEED,
        deterministic_trainer: bool = defaults.DETERMINISTIC_TRAINER,
        gradient_clip_val: Optional[float] = defaults.GRADIENT_CLIP_VAL,
        gradient_clip_algorithm: Optional[str] = defaults.GRADIENT_CLIP_ALGORITHM,
        mask_unused_classes: bool = defaults.MASK_UNUSED_CLASSES,
    ):
        learner_kwargs = {
            "memory_size": memory_size,
            "batch_memory_frac": batch_memory_frac,
            "loss_weight": loss_weight,
            "ema_memory_update_gamma": ema_memory_update_gamma,
            "loss_normalization": loss_normalization,
            "alpha": alpha,
            "beta": beta,
            "batch_size": batch_size,
            "seed": seed,
        }
        super().__init__(
            model,
            loss_fn=loss_fn,
            optimizer=optimizer,
            learner_class=DarkExperienceReplayLearner,
            learner_kwargs=learner_kwargs,
            input_state_folder=input_state_folder,
            output_state_folder=output_state_folder,
            max_epochs=max_epochs,
            learning_rate_scheduler=learning_rate_scheduler,
            learning_rate_scheduler_interval=learning_rate_scheduler_interval,
            train_transform=train_transform,
            train_target_transform=train_target_transform,
            test_transform=test_transform,
            test_target_transform=test_target_transform,
            buffer_transform=buffer_transform,
            buffer_target_transform=buffer_target_transform,
            metric=metric,
            mode=mode,
            logged_metrics=logged_metrics,
            early_stopping_enabled=early_stopping_enabled,
            logger=logger,
            accelerator=accelerator,
            devices=devices,
            strategy=strategy,
            precision=precision,
            deterministic_trainer=deterministic_trainer,
            gradient_clip_algorithm=gradient_clip_algorithm,
            gradient_clip_val=gradient_clip_val,
            mask_unused_classes=mask_unused_classes,
        )


class PooledOutputDistillationExperienceReplayModelUpdater(SingleTrainingLoopUpdater):
    def __init__(
        self,
        model: RenateModule,
        loss_fn: torch.nn.Module,
        optimizer: Callable[[List[Parameter]], Optimizer],
        memory_size: int,
        batch_memory_frac: int = defaults.BATCH_MEMORY_FRAC,
        loss_weight: float = defaults.LOSS_WEIGHT,
        ema_memory_update_gamma: float = defaults.EMA_MEMORY_UPDATE_GAMMA,
        loss_normalization: int = defaults.LOSS_NORMALIZATION,
        alpha: float = defaults.POD_ALPHA,
        distillation_type: str = defaults.POD_DISTILLATION_TYPE,
        normalize: bool = defaults.POD_NORMALIZE,
        learning_rate_scheduler: Optional[partial] = None,
        learning_rate_scheduler_interval: defaults.SUPPORTED_LR_SCHEDULER_INTERVAL_TYPE = defaults.LR_SCHEDULER_INTERVAL,  # noqa: E501
        batch_size: int = defaults.BATCH_SIZE,
        input_state_folder: Optional[str] = None,
        output_state_folder: Optional[str] = None,
        max_epochs: int = defaults.MAX_EPOCHS,
        train_transform: Optional[Callable] = None,
        train_target_transform: Optional[Callable] = None,
        test_transform: Optional[Callable] = None,
        test_target_transform: Optional[Callable] = None,
        buffer_transform: Optional[Callable] = None,
        buffer_target_transform: Optional[Callable] = None,
        metric: Optional[str] = None,
        mode: defaults.SUPPORTED_TUNING_MODE_TYPE = "min",
        logged_metrics: Optional[Dict[str, torchmetrics.Metric]] = None,
        early_stopping_enabled: bool = False,
        logger: Logger = defaults.LOGGER(**defaults.LOGGER_KWARGS),
        accelerator: defaults.SUPPORTED_ACCELERATORS_TYPE = defaults.ACCELERATOR,
        devices: Optional[int] = None,
        strategy: str = defaults.DISTRIBUTED_STRATEGY,
        precision: str = defaults.PRECISION,
        seed: int = defaults.SEED,
        deterministic_trainer: bool = defaults.DETERMINISTIC_TRAINER,
        gradient_clip_val: Optional[float] = defaults.GRADIENT_CLIP_VAL,
        gradient_clip_algorithm: Optional[str] = defaults.GRADIENT_CLIP_ALGORITHM,
        mask_unused_classes: bool = defaults.MASK_UNUSED_CLASSES,
    ):
        learner_kwargs = {
            "memory_size": memory_size,
            "batch_memory_frac": batch_memory_frac,
            "loss_weight": loss_weight,
            "ema_memory_update_gamma": ema_memory_update_gamma,
            "loss_normalization": loss_normalization,
            "alpha": alpha,
            "distillation_type": distillation_type,
            "normalize": normalize,
            "batch_size": batch_size,
            "seed": seed,
        }
        super().__init__(
            model,
            loss_fn=loss_fn,
            optimizer=optimizer,
            learner_class=PooledOutputDistillationExperienceReplayLearner,
            learner_kwargs=learner_kwargs,
            input_state_folder=input_state_folder,
            output_state_folder=output_state_folder,
            max_epochs=max_epochs,
            learning_rate_scheduler=learning_rate_scheduler,
            learning_rate_scheduler_interval=learning_rate_scheduler_interval,
            train_transform=train_transform,
            train_target_transform=train_target_transform,
            test_transform=test_transform,
            test_target_transform=test_target_transform,
            buffer_transform=buffer_transform,
            buffer_target_transform=buffer_target_transform,
            metric=metric,
            mode=mode,
            logged_metrics=logged_metrics,
            early_stopping_enabled=early_stopping_enabled,
            logger=logger,
            accelerator=accelerator,
            devices=devices,
            strategy=strategy,
            precision=precision,
            deterministic_trainer=deterministic_trainer,
            gradient_clip_algorithm=gradient_clip_algorithm,
            gradient_clip_val=gradient_clip_val,
            mask_unused_classes=mask_unused_classes,
        )


class CLSExperienceReplayModelUpdater(SingleTrainingLoopUpdater):
    def __init__(
        self,
        model: RenateModule,
        loss_fn: torch.nn.Module,
        optimizer: Callable[[List[Parameter]], Optimizer],
        memory_size: int,
        batch_memory_frac: int = defaults.BATCH_MEMORY_FRAC,
        loss_weight: float = defaults.LOSS_WEIGHT,
        ema_memory_update_gamma: float = defaults.EMA_MEMORY_UPDATE_GAMMA,
        loss_normalization: int = defaults.LOSS_NORMALIZATION,
        alpha: float = defaults.CLS_ALPHA,
        beta: float = defaults.CLS_BETA,
        stable_model_update_weight: float = defaults.CLS_STABLE_MODEL_UPDATE_WEIGHT,
        plastic_model_update_weight: float = defaults.CLS_PLASTIC_MODEL_UPDATE_WEIGHT,
        stable_model_update_probability: float = defaults.CLS_STABLE_MODEL_UPDATE_PROBABILITY,
        plastic_model_update_probability: float = defaults.CLS_PLASTIC_MODEL_UPDATE_PROBABILITY,
        learning_rate_scheduler: Optional[partial] = None,
        learning_rate_scheduler_interval: defaults.SUPPORTED_LR_SCHEDULER_INTERVAL_TYPE = defaults.LR_SCHEDULER_INTERVAL,  # noqa: E501
        batch_size: int = defaults.BATCH_SIZE,
        input_state_folder: Optional[str] = None,
        output_state_folder: Optional[str] = None,
        max_epochs: int = defaults.MAX_EPOCHS,
        train_transform: Optional[Callable] = None,
        train_target_transform: Optional[Callable] = None,
        test_transform: Optional[Callable] = None,
        test_target_transform: Optional[Callable] = None,
        buffer_transform: Optional[Callable] = None,
        buffer_target_transform: Optional[Callable] = None,
        metric: Optional[str] = None,
        mode: defaults.SUPPORTED_TUNING_MODE_TYPE = "min",
        logged_metrics: Optional[Dict[str, torchmetrics.Metric]] = None,
        early_stopping_enabled: bool = False,
        logger: Logger = defaults.LOGGER(**defaults.LOGGER_KWARGS),
        accelerator: defaults.SUPPORTED_ACCELERATORS_TYPE = defaults.ACCELERATOR,
        devices: Optional[int] = None,
        strategy: str = defaults.DISTRIBUTED_STRATEGY,
        precision: str = defaults.PRECISION,
        seed: int = defaults.SEED,
        deterministic_trainer: bool = defaults.DETERMINISTIC_TRAINER,
        gradient_clip_val: Optional[float] = defaults.GRADIENT_CLIP_VAL,
        gradient_clip_algorithm: Optional[str] = defaults.GRADIENT_CLIP_ALGORITHM,
        mask_unused_classes: bool = defaults.MASK_UNUSED_CLASSES,
    ):
        learner_kwargs = {
            "memory_size": memory_size,
            "batch_memory_frac": batch_memory_frac,
            "loss_weight": loss_weight,
            "ema_memory_update_gamma": ema_memory_update_gamma,
            "loss_normalization": loss_normalization,
            "alpha": alpha,
            "beta": beta,
            "stable_model_update_weight": stable_model_update_weight,
            "plastic_model_update_weight": plastic_model_update_weight,
            "stable_model_update_probability": stable_model_update_probability,
            "plastic_model_update_probability": plastic_model_update_probability,
            "batch_size": batch_size,
            "seed": seed,
        }
        super().__init__(
            model,
            loss_fn=loss_fn,
            optimizer=optimizer,
            learner_class=CLSExperienceReplayLearner,
            learner_kwargs=learner_kwargs,
            input_state_folder=input_state_folder,
            output_state_folder=output_state_folder,
            max_epochs=max_epochs,
            learning_rate_scheduler=learning_rate_scheduler,
            learning_rate_scheduler_interval=learning_rate_scheduler_interval,
            train_transform=train_transform,
            train_target_transform=train_target_transform,
            test_transform=test_transform,
            test_target_transform=test_target_transform,
            buffer_transform=buffer_transform,
            buffer_target_transform=buffer_target_transform,
            metric=metric,
            mode=mode,
            logged_metrics=logged_metrics,
            early_stopping_enabled=early_stopping_enabled,
            logger=logger,
            accelerator=accelerator,
            devices=devices,
            strategy=strategy,
            precision=precision,
            deterministic_trainer=deterministic_trainer,
            gradient_clip_algorithm=gradient_clip_algorithm,
            gradient_clip_val=gradient_clip_val,
            mask_unused_classes=mask_unused_classes,
        )


class SuperExperienceReplayModelUpdater(SingleTrainingLoopUpdater):
    def __init__(
        self,
        model: RenateModule,
        loss_fn: torch.nn.Module,
        optimizer: Callable[[List[Parameter]], Optimizer],
        memory_size: int,
        batch_memory_frac: int = defaults.BATCH_MEMORY_FRAC,
        loss_weight: float = defaults.LOSS_WEIGHT,
        ema_memory_update_gamma: float = defaults.EMA_MEMORY_UPDATE_GAMMA,
        loss_normalization: int = defaults.LOSS_NORMALIZATION,
        der_alpha: float = defaults.SER_DER_ALPHA,
        der_beta: float = defaults.SER_DER_BETA,
        sp_shrink_factor: float = defaults.SER_SP_SHRINK_FACTOR,
        sp_sigma: float = defaults.SER_SP_SIGMA,
        cls_alpha: float = defaults.SER_CLS_ALPHA,
        cls_stable_model_update_weight: float = defaults.SER_CLS_STABLE_MODEL_UPDATE_WEIGHT,
        cls_plastic_model_update_weight: float = defaults.SER_CLS_PLASTIC_MODEL_UPDATE_WEIGHT,
        cls_stable_model_update_probability: float = defaults.SER_CLS_STABLE_MODEL_UPDATE_PROBABILITY,  # noqa: E501
        cls_plastic_model_update_probability: float = defaults.SER_CLS_PLASTIC_MODEL_UPDATE_PROBABILITY,  # noqa: E501
        pod_alpha: float = defaults.SER_POD_ALPHA,
        pod_distillation_type: str = defaults.SER_POD_DISTILLATION_TYPE,
        pod_normalize: bool = defaults.SER_POD_NORMALIZE,
        learning_rate_scheduler: Optional[partial] = None,
        learning_rate_scheduler_interval: defaults.SUPPORTED_LR_SCHEDULER_INTERVAL_TYPE = defaults.LR_SCHEDULER_INTERVAL,  # noqa: E501
        batch_size: int = defaults.BATCH_SIZE,
        input_state_folder: Optional[str] = None,
        output_state_folder: Optional[str] = None,
        max_epochs: int = defaults.MAX_EPOCHS,
        train_transform: Optional[Callable] = None,
        train_target_transform: Optional[Callable] = None,
        test_transform: Optional[Callable] = None,
        test_target_transform: Optional[Callable] = None,
        buffer_transform: Optional[Callable] = None,
        buffer_target_transform: Optional[Callable] = None,
        metric: Optional[str] = None,
        mode: defaults.SUPPORTED_TUNING_MODE_TYPE = "min",
        logged_metrics: Optional[Dict[str, torchmetrics.Metric]] = None,
        early_stopping_enabled: bool = False,
        logger: Logger = defaults.LOGGER(**defaults.LOGGER_KWARGS),
        accelerator: defaults.SUPPORTED_ACCELERATORS_TYPE = defaults.ACCELERATOR,
        devices: Optional[int] = None,
        strategy: str = defaults.DISTRIBUTED_STRATEGY,
        precision: str = defaults.PRECISION,
        seed: int = defaults.SEED,
        deterministic_trainer: bool = defaults.DETERMINISTIC_TRAINER,
        gradient_clip_val: Optional[float] = defaults.GRADIENT_CLIP_VAL,
        gradient_clip_algorithm: Optional[str] = defaults.GRADIENT_CLIP_ALGORITHM,
        mask_unused_classes: bool = defaults.MASK_UNUSED_CLASSES,
    ):
        learner_kwargs = {
            "memory_size": memory_size,
            "batch_memory_frac": batch_memory_frac,
            "loss_weight": loss_weight,
            "ema_memory_update_gamma": ema_memory_update_gamma,
            "loss_normalization": loss_normalization,
            "der_alpha": der_alpha,
            "der_beta": der_beta,
            "sp_shrink_factor": sp_shrink_factor,
            "sp_sigma": sp_sigma,
            "cls_alpha": cls_alpha,
            "cls_stable_model_update_weight": cls_stable_model_update_weight,
            "cls_plastic_model_update_weight": cls_plastic_model_update_weight,
            "cls_stable_model_update_probability": cls_stable_model_update_probability,
            "cls_plastic_model_update_probability": cls_plastic_model_update_probability,
            "pod_alpha": pod_alpha,
            "pod_distillation_type": pod_distillation_type,
            "pod_normalize": pod_normalize,
            "batch_size": batch_size,
            "seed": seed,
        }
        super().__init__(
            model,
            loss_fn=loss_fn,
            optimizer=optimizer,
            learner_class=SuperExperienceReplayLearner,
            learner_kwargs=learner_kwargs,
            input_state_folder=input_state_folder,
            output_state_folder=output_state_folder,
            max_epochs=max_epochs,
            learning_rate_scheduler=learning_rate_scheduler,
            learning_rate_scheduler_interval=learning_rate_scheduler_interval,
            train_transform=train_transform,
            train_target_transform=train_target_transform,
            test_transform=test_transform,
            test_target_transform=test_target_transform,
            buffer_transform=buffer_transform,
            buffer_target_transform=buffer_target_transform,
            metric=metric,
            mode=mode,
            logged_metrics=logged_metrics,
            early_stopping_enabled=early_stopping_enabled,
            logger=logger,
            accelerator=accelerator,
            devices=devices,
            strategy=strategy,
            precision=precision,
            deterministic_trainer=deterministic_trainer,
            gradient_clip_algorithm=gradient_clip_algorithm,
            gradient_clip_val=gradient_clip_val,
            mask_unused_classes=mask_unused_classes,
        )



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
import argparse
import json
import os
import subprocess
from pathlib import Path

if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--test-file",
        type=str,
        required=True,
        help="Test suite to run.",
    )
    parser.add_argument(
        "--seed",
        type=int,
        required=True,
        help="Seed.",
    )
    parser.add_argument(
        f"--requirements-file",
        type=str,
        required=False,
        help="Path to requirements file",
    )
    args = parser.parse_args()
    test_suite = "main"
    current_folder = Path(os.path.dirname(__file__))
    configs_folder = current_folder / "configs"
    test_file = configs_folder / "suites" / test_suite / args.test_file
    if not test_file.is_file():
        raise FileNotFoundError(f"Unknown test file '{test_file}'.")
    with open(test_file) as f:
        test_config = json.load(f)
    job_name = f"{test_config['job_name']}"
    requirements_file = args.requirements_file
    if not requirements_file:
        requirements_file = current_folder.parent.parent / "requirements.txt"
    process = subprocess.Popen(
        [
            "python",
            current_folder / "run_experiment.py",
            "--scenario-file",
            configs_folder / "scenarios" / test_config["scenario"],
            "--model-file",
            configs_folder / "models" / test_config["model"],
            "--updater-file",
            configs_folder / "updaters" / test_config["updater"],
            "--dataset-file",
            configs_folder / "datasets" / test_config["dataset"],
            "--backend",
            test_config["backend"],
            "--job-name",
            job_name,
            "--test-suite",
            test_suite,
            "--seed",
            str(args.seed),
            "--requirements-file",
            requirements_file,
        ]
    )
    process.wait()



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
from functools import partial
from typing import Any, Callable, Dict, List, Optional, Tuple

import torch
import torchmetrics
from pytorch_lightning.loggers.logger import Logger
from pytorch_lightning.utilities.types import STEP_OUTPUT
from torch.nn import Parameter
from torch.optim import Optimizer
from torch.utils.data import ConcatDataset, DataLoader, Dataset

from renate import defaults
from renate.memory import ReservoirBuffer
from renate.models import RenateModule
from renate.types import NestedTensors
from renate.updaters.learner import ReplayLearner
from renate.updaters.model_updater import SingleTrainingLoopUpdater
from renate.utils.misc import maybe_populate_mask_and_ignore_logits
from renate.utils.pytorch import ConcatRandomSampler


class OfflineExperienceReplayLearner(ReplayLearner):
    """Experience Replay in the offline version.

    The model will be trained on weighted mixture of losses computed on the new data and a replay
    buffer. In contrast to the online version, the buffer will only be updated after training has
    terminated.

    Args:
        loss_weight_new_data: The training loss will be a convex combination of the loss on the new
            data and the loss on the memory data. If a float (needs to be in [0, 1]) is given here,
            it will be used as the weight for the new data. If `None`, the weight will be set
            dynamically to `N_t / sum([N_1, ..., N_t])`, where `N_i` denotes the size of task/chunk
            `i` and the current task is `t`.
    """

    def __init__(self, loss_weight_new_data: Optional[float] = None, **kwargs) -> None:
        super().__init__(**kwargs)
        if loss_weight_new_data is not None and not (0.0 <= loss_weight_new_data <= 1.0):
            raise ValueError(
                "Value of loss_weight_new_data needs to be between 0 and 1,"
                f"got {loss_weight_new_data}."
            )
        self._loss_weight_new_data = loss_weight_new_data
        self._num_points_previous_tasks: int = 0

    def _create_metrics_collections(
        self, logged_metrics: Optional[Dict[str, torchmetrics.Metric]] = None
    ) -> None:
        super()._create_metrics_collections(logged_metrics)
        self._loss_collections["train_losses"]["memory_loss"] = torchmetrics.MeanMetric()

    def on_model_update_start(
        self,
        train_dataset: Dataset,
        val_dataset: Dataset,
        train_dataset_collate_fn: Optional[Callable] = None,
        val_dataset_collate_fn: Optional[Callable] = None,
        task_id: Optional[str] = None,
    ) -> None:
        """Called before a model update starts."""
        super().on_model_update_start(
            train_dataset=train_dataset,
            val_dataset=val_dataset,
            train_dataset_collate_fn=train_dataset_collate_fn,
            val_dataset_collate_fn=val_dataset_collate_fn,
            task_id=task_id,
        )
        self._num_points_current_task = len(train_dataset)

    def train_dataloader(self) -> DataLoader:
        if len(self._memory_buffer) > self._memory_batch_size:
            train_buffer = ReservoirBuffer(
                max_size=self._num_points_current_task,
                seed=0,
                transform=self._train_transform,
                target_transform=self._train_target_transform,
            )
            train_buffer.update(self._train_dataset)
            return DataLoader(
                dataset=ConcatDataset([train_buffer, self._memory_buffer]),
                generator=self._rng,
                pin_memory=True,
                collate_fn=self._train_collate_fn,
                batch_sampler=ConcatRandomSampler(
                    [self._num_points_current_task, len(self._memory_buffer)],
                    [self._batch_size, self._memory_batch_size],
                    0,
                    generator=self._rng,
                ),
            )
        self._batch_size += self._memory_batch_size
        self._memory_batch_size = 0
        return super().train_dataloader()

    def on_model_update_end(self) -> None:
        """Called right before a model update terminates."""
        self._memory_buffer.update(self._train_dataset)
        self._num_points_previous_tasks += self._num_points_current_task
        self._num_points_current_task = -1

    def training_step(
        self, batch: Tuple[NestedTensors, Dict[str, Any]], batch_idx: int
    ) -> STEP_OUTPUT:
        """PyTorch Lightning function to return the training loss."""
        if self._loss_weight_new_data is None:
            alpha = self._num_points_current_task / (
                self._num_points_current_task + self._num_points_previous_tasks
            )
        else:
            alpha = self._loss_weight_new_data
        alpha = torch.tensor(alpha, device=next(self.parameters()).device)
        if self._memory_batch_size:
            (inputs, targets), _ = batch
        else:
            inputs, targets = batch
        outputs = self(inputs)

        outputs, self._class_mask = maybe_populate_mask_and_ignore_logits(
            self._mask_unused_classes, self._class_mask, self._classes_in_current_task, outputs
        )
        loss = self._loss_fn(outputs, targets)
        if self._memory_batch_size:
            loss_current = loss[: self._batch_size].mean()
            loss_memory = loss[self._batch_size :].mean()
            self._loss_collections["train_losses"]["base_loss"](loss_current)
            self._loss_collections["train_losses"]["memory_loss"](loss_memory)
            loss = alpha * loss_current + (1.0 - alpha) * loss_memory
        else:
            loss = loss.mean()
            self._loss_collections["train_losses"]["base_loss"](loss)
        self._update_metrics(outputs, targets, "train")
        return {"loss": loss}

    def on_save_checkpoint(self, checkpoint: Dict[str, Any]) -> None:
        super().on_save_checkpoint(checkpoint)
        checkpoint["num_points_previous_tasks"] = self._num_points_previous_tasks

    def on_load_checkpoint(self, checkpoint: Dict[str, Any]) -> None:
        super().on_load_checkpoint(checkpoint)
        self._num_points_previous_tasks = checkpoint["num_points_previous_tasks"]


class OfflineExperienceReplayModelUpdater(SingleTrainingLoopUpdater):
    def __init__(
        self,
        model: RenateModule,
        loss_fn: torch.nn.Module,
        optimizer: Callable[[List[Parameter]], Optimizer],
        memory_size: int,
        batch_memory_frac: int = defaults.BATCH_MEMORY_FRAC,
        loss_weight_new_data: Optional[float] = None,
        learning_rate_scheduler: Optional[partial] = None,
        learning_rate_scheduler_interval: defaults.SUPPORTED_LR_SCHEDULER_INTERVAL_TYPE = defaults.LR_SCHEDULER_INTERVAL,  # noqa: E501
        batch_size: int = defaults.BATCH_SIZE,
        input_state_folder: Optional[str] = None,
        output_state_folder: Optional[str] = None,
        max_epochs: int = defaults.MAX_EPOCHS,
        train_transform: Optional[Callable] = None,
        train_target_transform: Optional[Callable] = None,
        test_transform: Optional[Callable] = None,
        test_target_transform: Optional[Callable] = None,
        buffer_transform: Optional[Callable] = None,
        buffer_target_transform: Optional[Callable] = None,
        metric: Optional[str] = None,
        mode: defaults.SUPPORTED_TUNING_MODE_TYPE = "min",
        logged_metrics: Optional[Dict[str, torchmetrics.Metric]] = None,
        early_stopping_enabled: bool = False,
        logger: Logger = defaults.LOGGER(**defaults.LOGGER_KWARGS),
        accelerator: defaults.SUPPORTED_ACCELERATORS_TYPE = defaults.ACCELERATOR,
        devices: Optional[int] = None,
        strategy: str = defaults.DISTRIBUTED_STRATEGY,
        precision: str = defaults.PRECISION,
        seed: int = defaults.SEED,
        deterministic_trainer: bool = defaults.DETERMINISTIC_TRAINER,
        gradient_clip_val: Optional[float] = defaults.GRADIENT_CLIP_VAL,
        gradient_clip_algorithm: Optional[str] = defaults.GRADIENT_CLIP_ALGORITHM,
        mask_unused_classes: bool = defaults.MASK_UNUSED_CLASSES,
    ):
        learner_kwargs = {
            "memory_size": memory_size,
            "batch_memory_frac": batch_memory_frac,
            "loss_weight_new_data": loss_weight_new_data,
            "batch_size": batch_size,
            "seed": seed,
        }
        super().__init__(
            model,
            loss_fn=loss_fn,
            optimizer=optimizer,
            learner_class=OfflineExperienceReplayLearner,
            learner_kwargs=learner_kwargs,
            input_state_folder=input_state_folder,
            output_state_folder=output_state_folder,
            max_epochs=max_epochs,
            learning_rate_scheduler=learning_rate_scheduler,
            learning_rate_scheduler_interval=learning_rate_scheduler_interval,
            train_transform=train_transform,
            train_target_transform=train_target_transform,
            test_transform=test_transform,
            test_target_transform=test_target_transform,
            buffer_transform=buffer_transform,
            buffer_target_transform=buffer_target_transform,
            metric=metric,
            mode=mode,
            logged_metrics=logged_metrics,
            early_stopping_enabled=early_stopping_enabled,
            logger=logger,
            accelerator=accelerator,
            devices=devices,
            strategy=strategy,
            precision=precision,
            deterministic_trainer=deterministic_trainer,
            gradient_clip_algorithm=gradient_clip_algorithm,
            gradient_clip_val=gradient_clip_val,
            mask_unused_classes=mask_unused_classes,
        )



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
from typing import Any, Callable, Dict, List, Optional

import torch
import torch.nn as nn
import torchmetrics
from pytorch_lightning.loggers.logger import Logger
from torch.nn import Parameter
from torch.optim import Optimizer
from torch.optim.lr_scheduler import _LRScheduler
from torch.utils.data import Dataset

from renate import defaults
from renate.benchmark.models.spromptmodel import SPromptTransformer
from renate.models import RenateModule
from renate.updaters.learner import Learner
from renate.updaters.model_updater import SingleTrainingLoopUpdater


class SPeftLearner(Learner):
    """Learner to implement S-Prompts from
        ```Wang, Yabin, et.al .
        "S-prompts learning with pre-trained transformers: An occam’s razor for domain incremental learning."  # noqa: E501
        Advances in Neural Information Processing Systems 35 (2022): 5682-5695.```


    Args:
        model: The SPromptTransformer model to be trained.
        loss_fn: Loss function to be trained with.
        optimizer: Partial optimizer used to create an optimizer by passing the model parameters.
        learning_rate_scheduler: Partial object of learning rate scheduler that will be created by
            passing the optimizer.
        learning_rate_scheduler_interval: When to update the learning rate scheduler.
            Options: `epoch` and `step`.
        batch_size: Training batch size.
        train_transform: The transformation applied during training.
        train_target_transform: The target transformation applied during testing.
        test_transform: The transformation at test time.
        test_target_transform: The target transformation at test time.
        logged_metrics: Metrics logged additional to the default ones.
        seed: See :func:`renate.models.utils.get_generator`.
        mask_unused_classes: Masking logits corresponding to unused classes. Useful only for class
            incremental problems. Defaults to defaults.MASK_UNUSED_CLASSES.
    """

    def __init__(
        self,
        model: RenateModule,
        loss_fn: torch.nn.Module,
        optimizer: Callable[[List[Parameter]], Optimizer],
        learning_rate_scheduler: Optional[Optional[Callable[[Optimizer], _LRScheduler]]] = None,
        learning_rate_scheduler_interval: defaults.SUPPORTED_LR_SCHEDULER_INTERVAL_TYPE = defaults.LR_SCHEDULER_INTERVAL,  # noqa: E501
        batch_size: int = defaults.BATCH_SIZE,
        train_transform: Optional[Callable] = None,
        train_target_transform: Optional[Callable] = None,
        test_transform: Optional[Callable] = None,
        test_target_transform: Optional[Callable] = None,
        logged_metrics: Optional[Dict[str, torchmetrics.Metric]] = None,
        seed: int = defaults.SEED,
        mask_unused_classes: bool = defaults.MASK_UNUSED_CLASSES,
    ) -> None:
        if not isinstance(model, SPromptTransformer):
            raise ValueError(
                "SPrompt Learner can only be used with a SPromptTransformer model."
                f"But got {type(model)}"
            )
        super().__init__(
            model,
            loss_fn,
            optimizer,
            learning_rate_scheduler,
            learning_rate_scheduler_interval,
            batch_size,
            train_transform,
            train_target_transform,
            test_transform,
            test_target_transform,
            logged_metrics,
            seed,
            mask_unused_classes,
        )

    def on_model_update_start(
        self,
        train_dataset: Dataset,
        val_dataset: Dataset,
        train_dataset_collate_fn: Optional[Callable] = None,
        val_dataset_collate_fn: Optional[Callable] = None,
        task_id: Optional[str] = None,
    ) -> None:
        """A custom on_model_update_start hook for S-Peft methods.

        Here, we iterate oer the train data set and extract features. These features used to compute
        the task prototypes by the `update_task_identifier` call. Having this function in the model
        update start instead of end results in val metrics being reflective of test accuracy.
        """
        super().on_model_update_start(
            train_dataset, val_dataset, train_dataset_collate_fn, val_dataset_collate_fn, task_id
        )
        ## k-means
        device = torch.device("cuda:0") if torch.cuda.is_available() else torch.device("cpu")
        self._model.to(device)
        features, labels = [], []
        with torch.inference_mode():
            for x, y in self.train_dataloader():
                features.append(self._model.features(x.to(device)).cpu())
                labels.append(y)
        features = torch.cat(features)
        labels = torch.cat(labels)
        self._model.update_task_identifier(features=features, labels=labels)

    def setup(self, stage: str) -> None:
        # We dont support distributed
        assert (
            self.trainer.world_size == 1
        ), "SPrompt learner does not support Multi-GPU training yet."
        if stage == "fit":
            # This needs to run before configure optimizers is called. The only hook is setup("fit")
            self._model.increment_task()

    def optimizer_zero_grad(
        self, epoch: int, batch_idx: int, optimizer: Optimizer, optimizer_idx: int
    ) -> None:
        """Explicitly setting grads to None instead of zero."""
        optimizer.zero_grad(set_to_none=True)


class SPeftModelUpdater(SingleTrainingLoopUpdater):
    def __init__(
        self,
        model: RenateModule,
        loss_fn: torch.nn.Module,
        optimizer: Callable[[List[nn.Parameter]], Optimizer],
        batch_size: int = defaults.BATCH_SIZE,
        seed: int = defaults.SEED,
        learner_kwargs: Optional[Dict[str, Any]] = None,
        input_state_folder: Optional[str] = None,
        output_state_folder: Optional[str] = None,
        max_epochs: int = defaults.MAX_EPOCHS,
        learning_rate_scheduler: Optional[Optional[Callable[[Optimizer], _LRScheduler]]] = None,
        learning_rate_scheduler_interval: defaults.SUPPORTED_LR_SCHEDULER_INTERVAL_TYPE = defaults.LR_SCHEDULER_INTERVAL,  # noqa: E501
        train_transform: Optional[Callable] = None,
        train_target_transform: Optional[Callable] = None,
        test_transform: Optional[Callable] = None,
        test_target_transform: Optional[Callable] = None,
        buffer_transform: Optional[Callable] = None,
        buffer_target_transform: Optional[Callable] = None,
        metric: Optional[str] = None,
        mode: defaults.SUPPORTED_TUNING_MODE_TYPE = "min",
        logged_metrics: Optional[Dict[str, torchmetrics.Metric]] = None,
        early_stopping_enabled: bool = False,
        logger: Logger = defaults.LOGGER(**defaults.LOGGER_KWARGS),
        accelerator: defaults.SUPPORTED_ACCELERATORS_TYPE = defaults.ACCELERATOR,
        devices: Optional[int] = None,
        strategy: Optional[str] = defaults.DISTRIBUTED_STRATEGY,
        precision: str = defaults.PRECISION,
        deterministic_trainer: bool = defaults.DETERMINISTIC_TRAINER,
        gradient_clip_val: Optional[float] = defaults.GRADIENT_CLIP_VAL,
        gradient_clip_algorithm: Optional[str] = defaults.GRADIENT_CLIP_ALGORITHM,
        mask_unused_classes: bool = defaults.MASK_UNUSED_CLASSES,
    ):
        learner_kwargs = {
            "batch_size": batch_size,
            "seed": seed,
            "loss_fn": loss_fn,
        }
        super().__init__(
            model=model,
            loss_fn=loss_fn,
            optimizer=optimizer,
            learner_class=SPeftLearner,
            learner_kwargs=learner_kwargs,
            input_state_folder=input_state_folder,
            output_state_folder=output_state_folder,
            max_epochs=max_epochs,
            learning_rate_scheduler=learning_rate_scheduler,
            learning_rate_scheduler_interval=learning_rate_scheduler_interval,
            train_transform=train_transform,
            train_target_transform=train_target_transform,
            test_transform=test_transform,
            test_target_transform=test_target_transform,
            buffer_transform=buffer_transform,
            buffer_target_transform=buffer_target_transform,
            metric=metric,
            mode=mode,
            logged_metrics=logged_metrics,
            early_stopping_enabled=early_stopping_enabled,
            logger=logger,
            accelerator=accelerator,
            devices=devices,
            strategy=strategy,
            precision=precision,
            deterministic_trainer=deterministic_trainer,
            gradient_clip_algorithm=gradient_clip_algorithm,
            gradient_clip_val=gradient_clip_val,
            mask_unused_classes=mask_unused_classes,
        )



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
import inspect
import json
import logging
import os
import shutil
import tempfile
import warnings
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple, Type, Union

import numpy as np
import pandas as pd
from sagemaker import get_execution_role
from sagemaker.pytorch import PyTorch
from syne_tune.backend.local_backend import LocalBackend
from syne_tune.config_space import Domain
from syne_tune.experiments import load_experiment
from syne_tune.optimizer.baselines import ASHA
from syne_tune.optimizer.scheduler import TrialScheduler
from syne_tune.optimizer.schedulers import FIFOScheduler
from syne_tune.optimizer.schedulers.transfer_learning import (
    RUSHScheduler,
    TransferLearningTaskEvaluations,
)
from syne_tune.results_callback import StoreResultsCallback
from syne_tune.stopping_criterion import StoppingCriterion
from syne_tune.tuner import Tuner
from syne_tune.util import experiment_path

import renate
from renate import defaults
from renate.cli.parsing_functions import get_data_module_fn_kwargs, to_dense_str
from renate.utils.file import move_to_uri
from renate.utils.module import get_and_prepare_data_module, import_module
from renate.utils.syne_tune import (
    TrainingLoggerCallback,
    TuningLoggerCallback,
    best_hyperparameters,
    config_space_to_dict,
    is_syne_tune_config_space,
    redirect_to_tmp,
)

logger = logging.getLogger(__name__)

RENATE_CONFIG_COLUMNS = [
    "config_file",
    "prepare_data",
    "chunk_id",
    "task_id",
    "working_directory",
    "seed",
    "accelerator",
    "devices",
    "metric",
    "mode",
    "input_state_url",
]


def run_training_job(
    mode: defaults.SUPPORTED_TUNING_MODE_TYPE,
    config_space: Dict[str, Any],
    metric: str,
    backend: defaults.SUPPORTED_BACKEND_TYPE,
    updater: str = defaults.LEARNER,
    max_epochs: int = defaults.MAX_EPOCHS,
    task_id: str = defaults.TASK_ID,
    chunk_id: Optional[int] = None,
    input_state_url: Optional[str] = None,
    output_state_url: Optional[str] = None,
    working_directory: Optional[str] = defaults.WORKING_DIRECTORY,
    dependencies: Optional[List[str]] = None,
    config_file: Optional[str] = None,
    requirements_file: Optional[str] = None,
    role: Optional[str] = None,
    instance_type: str = defaults.INSTANCE_TYPE,
    instance_count: int = defaults.INSTANCE_COUNT,
    instance_max_time: float = defaults.INSTANCE_MAX_TIME,
    max_time: Optional[float] = None,
    max_num_trials_started: Optional[int] = None,
    max_num_trials_completed: Optional[int] = None,
    max_num_trials_finished: Optional[int] = None,
    max_cost: Optional[float] = None,
    n_workers: int = defaults.N_WORKERS,
    scheduler: Optional[Union[str, Type[TrialScheduler]]] = None,
    scheduler_kwargs: Optional[Dict] = None,
    seed: int = defaults.SEED,
    accelerator: defaults.SUPPORTED_ACCELERATORS_TYPE = defaults.ACCELERATOR,
    devices: int = defaults.DEVICES,
    strategy: str = defaults.DISTRIBUTED_STRATEGY,
    precision: str = defaults.PRECISION,
    deterministic_trainer: bool = defaults.DETERMINISTIC_TRAINER,
    gradient_clip_val: Optional[float] = defaults.GRADIENT_CLIP_VAL,
    gradient_clip_algorithm: Optional[str] = defaults.GRADIENT_CLIP_ALGORITHM,
    job_name: str = defaults.JOB_NAME,
) -> Optional[Tuner]:
    """Starts updating the model including hyperparameter optimization.

    Args:
        mode: Declares the type of optimization problem: `min` or `max`.
        config_space: Details for defining your own search space is provided in the
            `Syne Tune Documentation
            <https://syne-tune.readthedocs.io/en/latest/search_space.html>`_.
        metric: Name of metric to optimize.
        backend: Whether to run jobs locally (`local`) or on SageMaker (`sagemaker`).
        updater: Updater used for model update.
        max_epochs: The maximum number of epochs used to train the model. For comparability between
            methods, epochs are interpreted as "finetuning-equivalent". That is, one epoch is
            defined as `len(current_task_dataset) / batch_size` update steps.
        task_id: Unique identifier for the current task.
        chunk_id: Unique identifier for the current data chunk.
        input_state_url: Path to the Renate model state.
        output_state_url: Path where Renate model state will be stored.
        working_directory: Path to the working directory.
        dependencies: (SageMaker backend only) List of strings containing absolute or relative paths
            to files and directories that will be uploaded as part of the SageMaker training job.
        config_file: File containing the definition of `model_fn` and `data_module_fn`.
        requirements_file: (SageMaker backend only) Path to requirements.txt containing environment
            dependencies.
        role: (SageMaker backend only) An AWS IAM role (either name or full ARN).
        instance_type: (SageMaker backend only) Sagemaker instance type for each worker.
        instance_count: (SageMaker backend only) Number of instances for each worker.
        instance_max_time: (SageMaker backend only) Requested maximum wall_clock time for each
            worker.
        max_time: Stopping criterion: wall clock time.
        max_num_trials_started: Stopping criterion: trials started.
        max_num_trials_completed: Stopping criterion: trials completed.
        max_num_trials_finished: Stopping criterion: trials finished.
        max_cost: (SageMaker backend only) Stopping criterion: SageMaker cost.
        n_workers: Number of workers running in parallel.
        scheduler: Default is random search, you can change it by providing either a string
            (`random`, `bo`, `asha` or `rush`) or scheduler class and its corresponding
            `scheduler_kwargs` if required. For latter option,
            `see details at <https://github.com/awslabs/syne-tune/blob/main/docs/schedulers.md>`_ .
        scheduler_kwargs: Only required if custom scheduler is provided.
        seed: Seed used for ensuring reproducibility.
        accelerator: Type of accelerator to use.
        devices: Number of devices to use per worker (set in n_workers).
        strategy: Name of the distributed training strategy to use.
            `More details <https://lightning.ai/docs/pytorch/stable/extensions/strategy.html>`__
        precision: Type of bit precision to use.
            `More details <https://lightning.ai/docs/pytorch/stable/common/precision_basic.html>`__
        deterministic_trainer: When true the Trainer adopts a deterministic behaviour also on GPU.
        gradient_clip_val: The value at which to clip gradients. Passing None disables it.
            `More details <https://lightning.ai/docs/pytorch/stable/common/trainer.html#init>`__
        gradient_clip_algorithm: The gradient clipping algorithm to use. Can be norm or value.
            `More details <https://lightning.ai/docs/pytorch/stable/common/trainer.html#init>`__
        job_name: Prefix for the name of the SageMaker training job.
    """
    assert (
        mode in defaults.SUPPORTED_TUNING_MODE
    ), f"Mode {mode} is not in {defaults.SUPPORTED_TUNING_MODE}."
    assert (
        backend in defaults.SUPPORTED_BACKEND
    ), f"Backend {backend} is not in {defaults.SUPPORTED_BACKEND}."
    for key, value in config_space.items():
        if isinstance(value, (bool, list, tuple)):
            config_space[key] = to_dense_str(value)
    if backend == "local":
        return _execute_training_and_tuning_job_locally(
            input_state_url=input_state_url,
            output_state_url=output_state_url,
            working_directory=working_directory,
            config_file=config_file,
            mode=mode,
            config_space=config_space,
            metric=metric,
            updater=updater,
            max_epochs=max_epochs,
            task_id=task_id,
            chunk_id=chunk_id,
            max_time=max_time,
            max_num_trials_started=max_num_trials_started,
            max_num_trials_completed=max_num_trials_completed,
            max_num_trials_finished=max_num_trials_finished,
            max_cost=max_cost,
            n_workers=n_workers,
            scheduler=scheduler,
            scheduler_kwargs=scheduler_kwargs,
            seed=seed,
            accelerator=accelerator,
            devices=devices,
            strategy=strategy,
            precision=precision,
            gradient_clip_algorithm=gradient_clip_algorithm,
            gradient_clip_val=gradient_clip_val,
            deterministic_trainer=deterministic_trainer,
        )
    submit_remote_job(
        input_state_url=input_state_url,
        output_state_url=output_state_url,
        working_directory=working_directory,
        config_file=config_file,
        mode=mode,
        config_space=config_space,
        metric=metric,
        updater=updater,
        max_epochs=max_epochs,
        task_id=task_id,
        chunk_id=chunk_id,
        dependencies=dependencies or [],
        requirements_file=requirements_file,
        role=role,
        instance_type=instance_type,
        instance_count=instance_count,
        instance_max_time=instance_max_time,
        max_time=max_time,
        max_num_trials_started=max_num_trials_started,
        max_num_trials_completed=max_num_trials_completed,
        max_num_trials_finished=max_num_trials_finished,
        max_cost=max_cost,
        n_workers=n_workers,
        scheduler=scheduler,
        scheduler_kwargs=scheduler_kwargs,
        seed=seed,
        accelerator=accelerator,
        devices=devices,
        strategy=strategy,
        precision=precision,
        deterministic_trainer=deterministic_trainer,
        gradient_clip_algorithm=gradient_clip_algorithm,
        gradient_clip_val=gradient_clip_val,
        job_name=job_name,
    )


def _prepare_remote_job(
    tmp_dir: str,
    requirements_file: Optional[str],
    optional_dependencies: Optional[str] = None,
    **job_kwargs: Any,
) -> List[str]:
    """Prepares a SageMaker job."""
    dependencies = list(renate.__path__ + [job_kwargs["config_file"]])

    if "input_state_url" in job_kwargs and job_kwargs["input_state_url"] is None:
        del job_kwargs["input_state_url"]
    job_kwargs["config_file"] = os.path.basename(job_kwargs["config_file"])
    job_kwargs["config_space"] = config_space_to_dict(job_kwargs["config_space"])

    jobs_kwargs_file = os.path.join(tmp_dir, defaults.JOB_KWARGS_FILE)
    with open(jobs_kwargs_file, "w") as f:
        json.dump(job_kwargs, f)
    dependencies.append(jobs_kwargs_file)

    if requirements_file is None:
        requirements_file = os.path.join(tmp_dir, "requirements.txt")
        with open(requirements_file, "w") as f:
            f.write(
                "Renate{}=={}".format(
                    "" if optional_dependencies is None else f"[{optional_dependencies}]",
                    renate.__version__,
                )
            )
    dependencies.append(requirements_file)
    return dependencies


def _get_transfer_learning_task_evaluations(
    tuning_results: pd.DataFrame,
    config_space: Dict[str, Any],
    metric: str,
    max_epochs: int,
) -> Optional[TransferLearningTaskEvaluations]:
    """Converts data frame with training results of a single update step into
    `TransferLearningTaskEvaluations`.

    Args:
        tuning_results: Results of previous hyperparameter optimization runs.
        config_space: Configuration space used.
        metric: The metric to be optimized. This will be the only metric added to
            `TransferLearningTaskEvaluations`.
        max_epochs: Length of the learning curve. Learning curves will be padded with `np.nan`
            to this length.
    Returns:
        `TransferLearningTaskEvaluations` contains a `pd.DataFrame` of hyperparameters, a `np.array`
        of shape `(num_hyperparameters, 1, max_epochs, 1)` which contains the learning curves for
        `metric` for each hyperparameter.
    """
    for config in RENATE_CONFIG_COLUMNS:
        config_with_prefix = f"config_{config}"
        if config in config_space:
            tuning_results[config_with_prefix] = config_space[config]
        elif config_with_prefix in tuning_results:
            del tuning_results[config_with_prefix]
    if (
        metric not in tuning_results
        or not len(tuning_results)
        or not set([f"config_{config_name}" for config_name in config_space]).issubset(
            tuning_results
        )
    ):
        return None

    def pad_same_length(x):
        if len(x) == max_epochs:
            return x
        return x + (max_epochs - len(x)) * [np.nan]

    hyperparameter_names = [
        hyperparameter for hyperparameter in tuning_results if hyperparameter.startswith("config_")
    ]
    hyperparameters = tuning_results[hyperparameter_names]
    hyperparameters.columns = [
        hyperparameter_name[7:] for hyperparameter_name in hyperparameters.columns
    ]
    hyperparameters = hyperparameters[list(config_space)]
    hyperparameter_rows_with_nan = hyperparameters.isna().any(axis=1)
    hyperparameter_duplicates = hyperparameters.duplicated()
    objectives_evaluations = np.array(tuning_results[metric].apply(pad_same_length).to_list())
    objectives_evaluations = np.expand_dims(objectives_evaluations, axis=(1, 3))
    objectives_evaluations_with_nan_only = np.all(np.isnan(objectives_evaluations), axis=2).reshape(
        -1
    )
    valid_hyperparameters = np.logical_and(
        np.logical_and(~hyperparameter_rows_with_nan, ~objectives_evaluations_with_nan_only),
        ~hyperparameter_duplicates,
    )
    if not np.any(valid_hyperparameters):
        return None
    hyperparameters = hyperparameters[valid_hyperparameters]
    objectives_evaluations = objectives_evaluations[valid_hyperparameters]
    return TransferLearningTaskEvaluations(
        configuration_space=config_space,
        hyperparameters=hyperparameters,
        objectives_names=[metric],
        objectives_evaluations=objectives_evaluations,
    )


def _load_tuning_history(
    input_state_url: str, config_space: Dict[str, Any], metric: str
) -> Dict[str, TransferLearningTaskEvaluations]:
    """Loads the tuning history in a list where each entry of the list is the tuning history of one
    update.

    Args:
        input_state_url: Location of state. Will check at this location of a tuning history exists.
        config_space: The configuration space defines which parts of the tuning history to load.
        metric: Only the defined metric of the tuning history will be loaded.
    Returns:
        Returns an empty list if no previous tuning history exists or it does not match the current
        `config_space`. The list contains an instance of `TransferLearningTaskEvaluations` for each
        update that contains matching data.
    """
    if input_state_url is None or not Path(defaults.hpo_file(input_state_url)).exists():
        return {}
    tuning_results = pd.read_csv(defaults.hpo_file(input_state_url))
    hyperparameter_names = [
        hyperparameter for hyperparameter in tuning_results if hyperparameter.startswith("config_")
    ]
    agg_dict = {
        hyperparameter_name: lambda x: x.iloc[0] for hyperparameter_name in hyperparameter_names
    }
    agg_dict[metric] = list
    max_epochs = tuning_results["epoch"].max()
    task_evaluations = {}
    for update_id in tuning_results["update_id"].unique():
        task_evaluation = _get_transfer_learning_task_evaluations(
            tuning_results=tuning_results[tuning_results["update_id"] == update_id]
            .groupby(["trial_id"])
            .agg(agg_dict),
            config_space=config_space,
            metric=metric,
            max_epochs=max_epochs,
        )
        if task_evaluation is not None:
            task_evaluations[f"update_id_{update_id}"] = task_evaluation
    return task_evaluations


def _merge_tuning_history(
    new_tuning_results: pd.DataFrame, old_tuning_results: Optional[pd.DataFrame] = None
) -> pd.DataFrame:
    """Merges old tuning history with training results from current chunk.

    `update_id` identifies the update step in the csv file. This allows creating the metadata
        required for transfer hyperparameter optimization.
    """
    update_id = 0 if old_tuning_results is None else old_tuning_results["update_id"].max() + 1
    new_tuning_results.insert(0, "update_id", update_id)
    if old_tuning_results is not None:
        return pd.concat([old_tuning_results, new_tuning_results], axis=0, ignore_index=True)
    return new_tuning_results


def _teardown_tuning_job(
    backend: LocalBackend,
    config_space: Dict[str, Union[Domain, int, float, str]],
    job_name: str,
    input_state_url: Optional[str] = None,
    output_state_url: Optional[str] = None,
) -> None:
    """Update lifelong hyperparameter optimization results, save state and clean up disk."""
    experiment_folder = redirect_to_tmp(str(experiment_path(job_name)))
    if output_state_url is not None:
        experiment = load_experiment(job_name)
        try:
            best_trial_id = experiment.best_config()["trial_id"]
            if is_syne_tune_config_space(config_space):
                logger.info(
                    "Best hyperparameter settings: "
                    f"{best_hyperparameters(experiment, config_space)}"
                )
        except AttributeError:
            raise RuntimeError(
                "Not a single training run finished. This may have two reasons:\n"
                "1) The provided tuning time is too short.\n"
                "2) There is a bug in the training script."
                + "\n\nLogs (stdout):\n\n{}".format("".join(backend.stdout(0)))
                + "\n\nLogs (stderr):\n\n{}".format("".join(backend.stderr(0)))
            )
        output_state_folder = defaults.output_state_folder(
            f"{experiment_folder}/{best_trial_id}/checkpoints"
        )
        old_tuning_results = (
            pd.read_csv(defaults.hpo_file(input_state_url))
            if input_state_url is not None and os.path.exists(defaults.hpo_file(input_state_url))
            else None
        )
        tuning_results = _merge_tuning_history(experiment.results, old_tuning_results)
        tuning_results.to_csv(defaults.hpo_file(output_state_folder), index=False)
        move_to_uri(output_state_folder, output_state_url)
        logger.info(f"Renate state is available at {output_state_url}.")
    shutil.rmtree(experiment_folder, ignore_errors=True)
    shutil.rmtree(experiment_path(job_name), ignore_errors=True)


def _verify_validation_set_for_hpo_and_checkpointing(
    config_space: Dict[str, Any],
    config_file: str,
    tune_hyperparameters: bool,
    metric: str,
    mode: defaults.SUPPORTED_TUNING_MODE_TYPE,
    working_directory: str,
) -> Tuple[str, defaults.SUPPORTED_TUNING_MODE_TYPE]:
    """Checks if validation set is provided when needed and updates config_space such that
    checkpointing works.

    If a validation set exists, the metric is forwarded such that we store the checkpoint which
    performs best on validation. This is a side effect changing `config_space`. Otherwise, the
    checkpoint of the last epoch is used.

    Returns:
        Metric and mode used by the Syne Tune tuner. If there is no validation set, returns
        `("train_loss", "min")`.
    Raises:
        AssertionError: If `tune_hyperparameters` is True but no validation set is provided.
    """
    config_module = import_module("config_module", config_file)
    data_module = get_and_prepare_data_module(
        config_module,
        data_path=defaults.data_folder(working_directory),
        **get_data_module_fn_kwargs(config_module, config_space, cast_arguments=True),
    )
    data_module.setup()
    val_exists = data_module.val_data() is not None
    assert (
        val_exists or not tune_hyperparameters
    ), "Provide a validation set to optimize hyperparameters."
    if val_exists:
        config_space["metric"] = metric
        config_space["mode"] = mode
        return metric, mode
    elif not val_exists and metric is not None:
        warnings.warn("No need to pass `metric`. Without a validation set, it is not used.")
    return "train_loss", "min"


def _create_scheduler(
    scheduler: Union[str, Type[TrialScheduler]],
    config_space: Dict[str, Any],
    metric: str,
    mode: defaults.SUPPORTED_TUNING_MODE_TYPE,
    seed: int,
    scheduler_kwargs: Optional[Dict[str, Any]] = None,
    input_state_url: Optional[str] = None,
) -> TrialScheduler:
    scheduler_kwargs = scheduler_kwargs or {}
    if isinstance(scheduler, str):
        hyperband_scheduler_kwargs = {"max_resource_attr": "max_epochs", "resource_attr": "epoch"}
        scheduler_classes = {
            "asha": {"scheduler": ASHA, "scheduler_kwargs": hyperband_scheduler_kwargs},
            "bo": {"scheduler": FIFOScheduler, "scheduler_kwargs": {"searcher": "bayesopt"}},
            "rush": {"scheduler": RUSHScheduler, "scheduler_kwargs": hyperband_scheduler_kwargs},
            "random": {"scheduler": FIFOScheduler, "scheduler_kwargs": {"searcher": "random"}},
        }
        assert (
            scheduler in scheduler_classes
        ), f"Unknown scheduler {scheduler}. Options: {list(scheduler_classes)}."
        scheduler_kwargs.update(scheduler_classes[scheduler]["scheduler_kwargs"])
        scheduler = scheduler_classes[scheduler]["scheduler"]
    if "transfer_learning_evaluations" in inspect.getfullargspec(scheduler.__init__).args:
        scheduler_kwargs["transfer_learning_evaluations"] = _load_tuning_history(
            input_state_url=input_state_url, config_space=config_space, metric=metric
        )
        if scheduler_kwargs["transfer_learning_evaluations"]:
            logger.info(
                f"Using information of {len(scheduler_kwargs['transfer_learning_evaluations'])} "
                "previous tuning jobs to accelerate this job."
            )
    return scheduler(
        config_space=config_space,
        mode=mode,
        metric=metric,
        random_seed=seed,
        **scheduler_kwargs,
    )


def _execute_training_and_tuning_job_locally(
    input_state_url: Optional[str],
    output_state_url: Optional[str],
    working_directory: Optional[str],
    config_file: str,
    mode: defaults.SUPPORTED_TUNING_MODE_TYPE,
    config_space: Dict[str, Any],
    metric: str,
    updater: str,
    max_epochs: int,
    task_id: str,
    chunk_id: int,
    max_time: float,
    max_num_trials_started: int,
    max_num_trials_completed: int,
    max_num_trials_finished: int,
    max_cost: float,
    n_workers: int,
    scheduler: Union[str, Type[TrialScheduler]],
    scheduler_kwargs: Dict[str, Any],
    seed: int,
    accelerator: str,
    devices: int,
    deterministic_trainer: bool,
    strategy: str,
    precision: str,
    gradient_clip_algorithm: Optional[str],
    gradient_clip_val: Optional[float],
):
    """Executes the training job locally.

    See renate.train.run_training_job for a description of arguments.
    """
    tune_hyperparameters = is_syne_tune_config_space(config_space)
    config_space["updater"] = updater
    config_space["max_epochs"] = max_epochs
    config_space["config_file"] = config_file
    config_space["prepare_data"] = False
    if chunk_id is not None:
        config_space["chunk_id"] = chunk_id
    config_space["task_id"] = task_id
    config_space["working_directory"] = working_directory
    config_space["seed"] = seed
    config_space["accelerator"] = accelerator
    config_space["devices"] = devices
    config_space["strategy"] = strategy
    config_space["precision"] = precision
    config_space["deterministic_trainer"] = deterministic_trainer
    config_space["gradient_clip_val"] = gradient_clip_val
    config_space["gradient_clip_algorithm"] = gradient_clip_algorithm
    if input_state_url is not None:
        config_space["input_state_url"] = input_state_url

    metric, mode = _verify_validation_set_for_hpo_and_checkpointing(
        config_space=config_space,
        config_file=config_file,
        tune_hyperparameters=tune_hyperparameters,
        metric=metric,
        mode=mode,
        working_directory=working_directory,
    )

    training_script = str(Path(renate.__path__[0]) / "cli" / "run_training.py")
    assert Path(training_script).is_file(), f"Could not find training script {training_script}."
    logger.info("Start updating the model.")
    if tune_hyperparameters:
        logger.info(
            f"Tuning hyperparameters with respect to {metric} ({mode}) for {max_time} seconds on "
            f"{n_workers} worker(s)."
        )

    backend = LocalBackend(entry_point=training_script, num_gpus_per_trial=devices)
    if scheduler is None or not tune_hyperparameters:
        if scheduler is not None:
            warnings.warn(
                "Configuration space contains exactly one configuration, custom scheduler is "
                "ignored."
            )
        scheduler = defaults.scheduler(config_space=config_space, mode=mode, metric=metric)
    else:
        scheduler = _create_scheduler(
            scheduler=scheduler,
            config_space=config_space,
            metric=metric,
            mode=mode,
            seed=seed,
            scheduler_kwargs=scheduler_kwargs,
            input_state_url=input_state_url,
        )
    logging_callback = (
        TuningLoggerCallback(mode=mode, metric=metric)
        if tune_hyperparameters
        else TrainingLoggerCallback()
    )
    tuner = Tuner(
        trial_backend=backend,
        scheduler=scheduler,
        stop_criterion=StoppingCriterion(
            max_wallclock_time=max_time,
            max_num_trials_started=max_num_trials_started,
            max_num_trials_completed=max_num_trials_completed,
            max_num_trials_finished=max_num_trials_finished,
            max_cost=max_cost,
        ),
        n_workers=n_workers,
        callbacks=[StoreResultsCallback(), logging_callback],
    )

    try:
        tuner.run()
    except ValueError:
        raise RuntimeError(
            "Tuning failed."
            + "\n\nLogs (stdout):\n\n{}".format("".join(backend.stdout(0)))
            + "\n\nLogs (stderr):\n\n{}".format("".join(backend.stderr(0)))
        )

    logger.info("All training is completed. Saving state...")

    _teardown_tuning_job(
        backend=backend,
        config_space=config_space,
        job_name=tuner.name,
        input_state_url=input_state_url,
        output_state_url=output_state_url,
    )

    logger.info("Renate update completed successfully.")

    return tuner


def submit_remote_job(
    dependencies: List[str],
    role: str,
    instance_type: str,
    instance_count: int,
    instance_max_time: float,
    job_name: str,
    optional_dependencies: Optional[str] = None,
    **job_kwargs: Any,
) -> str:
    """Executes the training job on SageMaker.

    See renate.train.run_training_job for a description of arguments."""
    tuning_script = str(Path(renate.__path__[0]) / "cli" / "run_remote_job.py")
    job_timestamp = defaults.current_timestamp()
    job_name = f"{job_name}-{job_timestamp}"
    tmp_dir = tempfile.mkdtemp()
    dependencies += _prepare_remote_job(
        tmp_dir=tmp_dir, optional_dependencies=optional_dependencies, **job_kwargs
    )
    PyTorch(
        entry_point=tuning_script,
        instance_type=instance_type,
        instance_count=instance_count,
        py_version=defaults.PYTHON_VERSION,
        framework_version=defaults.FRAMEWORK_VERSION,
        max_run=instance_max_time,
        role=role or get_execution_role(),
        dependencies=dependencies,
        volume_size=defaults.VOLUME_SIZE,
    ).fit(wait=False, job_name=job_name)
    shutil.rmtree(tmp_dir)
    return job_name



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
from typing import Callable, Optional, Tuple

import torch
from torch.utils.data import Dataset

from renate import defaults
from renate.benchmark.datasets.base import DataIncrementalDataModule
from renate.data.data_module import RenateDataModule
from renate.utils.pytorch import get_generator


class DummyDataset(Dataset):
    def __init__(self, data, targets, transform=None):
        assert len(data) == len(targets)
        self.data = data
        self.targets = targets
        self.transform = transform

    def __getitem__(self, index):
        x = self.data[index]
        y = self.targets[index]
        if self.transform is not None:
            x = self.transform(x)
        return x, y

    def __len__(self):
        return len(self.data)


class DummyTorchVisionDataModule(RenateDataModule):
    """
    A simple data module similar to `TorchVisionDataModule` with 100 training instances and 100 test
    instances with shape (1, 5, 5) and 5 classes.
    """

    def __init__(
        self,
        transform: Optional[Callable] = None,
        val_size: float = defaults.VALIDATION_SIZE,
        seed: int = defaults.SEED,
    ):
        super().__init__(
            data_path="",
            src_bucket="",
            src_object_name="",
            val_size=val_size,
            seed=seed,
        )
        self._transform = transform
        self.input_shape = (1, 5, 5)
        self.output_shape = (1,)
        self.output_classes = 5

    def prepare_data(self) -> None:
        pass

    def _get_random_data(self, offset: int = 0):
        rng = get_generator(self._seed + offset)
        X = torch.randint(0, 42, (100, *self.input_shape), generator=rng, dtype=torch.float32)
        y = torch.ones(100, dtype=torch.long)
        for i in range(5):
            y[i * 20 : (i + 1) * 20] = i
        return X, y

    def setup(self):
        self.X_train, self.y_train = self._get_random_data()
        train_data = DummyDataset(self.X_train, self.y_train, self._transform)
        self._train_data, self._val_data = self._split_train_val_data(train_data)
        self.X_test, self.y_test = self._get_random_data()
        self._test_data = DummyDataset(self.X_test, self.y_test, self._transform)


class DummyDataIncrementalDataModule(DataIncrementalDataModule):
    """Simple dataset to test whether DataIncrementalDataModule or DataIncrementalScenario work.

    This data module gives access to 5 different datasets with ids [0, 1, 2, 3, 4].
    The labels of dataset with data_id=i are all i.
    """

    def __init__(
        self,
        data_id: int,
        input_shape: Tuple[int, ...],
        transform: Optional[Callable] = None,
        val_size: float = defaults.VALIDATION_SIZE,
        seed: int = defaults.SEED,
    ):
        super().__init__(data_path="", data_id=data_id, val_size=val_size, seed=seed)
        rng = get_generator(self._seed)

        def get_data(i):
            return DummyDataset(
                data=torch.rand((100, *input_shape), generator=rng, dtype=torch.float32),
                targets=torch.zeros(100) + i,
                transform=transform,
            )

        self._datasets = [[get_data(i) for _ in range(2)] for i in range(5)]

    def prepare_data(self) -> None:
        pass

    def setup(self):
        self._train_data, self._test_data = self._datasets[self.data_id]
        self._train_data, self._val_data = self._split_train_val_data(self._train_data)



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
from typing import Dict, Union

import numpy as np
from syne_tune.config_space import Domain, choice, loguniform, uniform


def _get_range(start, end, step):
    return [np.round(i, 8) for i in np.arange(start, end + 1e-8, step)]


_learner_config_space = {
    "optimizer": "SGD",
    "momentum": choice([0.0, 0.9, 0.99]),
    "weight_decay": loguniform(1e-6, 1e-2),
    "learning_rate": loguniform(0.001, 0.5),
    "batch_size": 64,
    "max_epochs": 50,
}
_replay_config_space = {
    **_learner_config_space,
    **{
        "batch_memory_frac": 0.5,
        "memory_size": 1000,
    },
}
_er_config_space = {
    **_replay_config_space,
    **{"updater": "ER", "alpha": 0.5, "loss_normalization": 0, "loss_weight": 0.5},
}
_der_config_space = {
    **_replay_config_space,
    **{
        "updater": "DER",
        "alpha": uniform(0.0, 1.0),
        "beta": uniform(0.0, 1.0),
        "loss_normalization": 0,
        "loss_weight": 1.0,
    },
}
_super_er_config_space = {
    **_replay_config_space,
    **{
        "updater": "Super-ER",
        "der_alpha": uniform(0.0, 1.0),
        "der_beta": uniform(0.0, 1.0),
        "sp_shrink_factor": choice(_get_range(0.3, 1, 0.1)),
        "sp_sigma": choice([0.0, 1e-5, 1e-4, 1e-3, 1e-2]),
        "cls_alpha": choice(_get_range(0, 1, 0.05)),
        "cls_stable_model_update_weight": 0.999,
        "cls_plastic_model_update_weight": 0.999,
        "cls_stable_model_update_probability": uniform(0.01, 0.49),
        "cls_plastic_model_update_probability": uniform(0.5, 1.0),
        "loss_normalization": 1,
        "ema_memory_update_gamma": uniform(0.95, 1.0),
    },
}
_repeated_distill_config_space = _replay_config_space
_offline_er_config_space = {
    **_replay_config_space,
    **{"loss_weight_new_data": choice([0.25, 0.5, 0.75])},
}


def config_space(updater: str) -> Dict[str, Union[Domain, str, int, float]]:
    """Returns the default configuration space for the updater."""
    config_spaces = {
        "ER": _er_config_space,
        "DER": _der_config_space,
        "SUPER-ER": _super_er_config_space,
        "RD": _repeated_distill_config_space,
        "OfflineER": _offline_er_config_space,
    }
    return config_spaces[updater.upper()]



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
import pytest
import torch

from renate.defaults import TASK_ID


def test_renate_mlp_init():
    pytest.helpers.get_renate_module_mlp(
        num_inputs=10, num_outputs=10, hidden_size=32, num_hidden_layers=3
    )


def test_renate_mlp_fwd():
    mlp = pytest.helpers.get_renate_module_mlp(
        num_inputs=10, num_outputs=10, hidden_size=32, num_hidden_layers=3
    )
    x = torch.rand(5, 10)
    y_hat = mlp(x)

    assert y_hat.shape[0] == 5
    assert y_hat.shape[1] == 10


@pytest.mark.parametrize("num_hidden_layers,expected_num_params", [[1, 4], [2, 6], [3, 8]])
def test_renate_mlp_get_params(num_hidden_layers, expected_num_params):
    mlp = pytest.helpers.get_renate_module_mlp(
        num_inputs=10, num_outputs=10, hidden_size=5, num_hidden_layers=num_hidden_layers
    )
    mlp.add_task_params(TASK_ID)

    first_task_params = mlp.get_params(TASK_ID)

    mlp.add_task_params("second_task")

    second_task_params = mlp.get_params("second_task")
    assert len(first_task_params) == len(second_task_params)
    # +2 for the output layer (weight and bias)
    # +4 for the 2 output layers  (weight and bias)
    assert len(first_task_params) == expected_num_params + 2
    assert len(second_task_params) == expected_num_params + 2
    assert len(list(mlp.parameters())) == expected_num_params + 4
    for i in range(len(first_task_params) - 2):
        # -2 because the last two parameters are weight and bias of a task specific linear layer
        assert torch.equal(first_task_params[i], second_task_params[i])

    for i in range(len(first_task_params) - 2, len(first_task_params)):
        # -2 because the last two parameters are weight and bias of a task specific linear layer
        assert not torch.equal(first_task_params[i], second_task_params[i])



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
from typing import List, Optional, Tuple, Union

import torch.nn as nn

from renate.benchmark.models.base import RenateBenchmarkingModule
from renate.models.prediction_strategies import PredictionStrategy


class MultiLayerPerceptron(RenateBenchmarkingModule):
    """A simple Multi Layer Perceptron with hidden layers, activation and Batch Normalization if
    enabled.

    Args:
        num_inputs: Number of input nodes.
        num_outputs: Number of output nodes.
        num_hidden_layers: Number of hidden layers.
        hidden_size: Uniform hidden size or the list or tuple of hidden sizes for individual hidden
            layers.
        activation: Activation name, matching activation name in `torch.nn` to be used between the
            hidden layers.
        batch_normalization: Whether to use Batch Normalization after the activation. By default the
            Batch Normalization tracks the running statistics.
        prediction_strategy: Continual learning strategies may alter the prediction at train or test
            time.
        add_icarl_class_means: If ``True``, additional parameters used only by the
            ``ICaRLModelUpdater`` are added. Only required when using that updater.
    """

    def __init__(
        self,
        num_inputs: int,
        num_outputs: int,
        num_hidden_layers: int,
        hidden_size: Union[int, List[int], Tuple[int]],
        activation: str = "ReLU",
        batch_normalization: bool = False,
        prediction_strategy: Optional[PredictionStrategy] = None,
        add_icarl_class_means: bool = True,
    ) -> None:
        embedding_size = hidden_size if type(hidden_size) == int else hidden_size[-1]
        super().__init__(
            embedding_size=embedding_size,
            num_outputs=num_outputs,
            constructor_arguments={
                "num_inputs": num_inputs,
                "num_hidden_layers": num_hidden_layers,
                "hidden_size": hidden_size,
                "activation": activation,
                "batch_normalization": batch_normalization,
            },
            prediction_strategy=prediction_strategy,
            add_icarl_class_means=add_icarl_class_means,
        )
        if isinstance(hidden_size, int):
            hidden_size = [hidden_size for _ in range(num_hidden_layers + 1)]
        assert len(hidden_size) == num_hidden_layers + 1

        activation = getattr(nn, activation)

        hidden_size = [num_inputs] + hidden_size
        layers = [nn.Flatten()]
        for i in range(num_hidden_layers + 1):
            layers.append(nn.Linear(hidden_size[i], hidden_size[i + 1]))
            layers.append(activation())
            if batch_normalization:
                layers.append(nn.BatchNorm1d(hidden_size[i + 1]))

        self._backbone = nn.Sequential(*layers)



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
from typing import Any, Optional, Tuple, Union

import torch
from transformers import ViTConfig, ViTModel
from transformers.modeling_outputs import BaseModelOutputWithPooling

from renate import defaults
from renate.benchmark.models.base import RenateBenchmarkingModule
from renate.models.prediction_strategies import PredictionStrategy


class FeatureExtractorViTModel(ViTModel):
    """This class directly outputs [CLS] features if cls_feat is True else returns per patch
    embeddings
    """

    def forward(
        self,
        pixel_values: Optional[torch.Tensor] = None,
        head_mask: Optional[torch.Tensor] = None,
        output_attentions: Optional[bool] = None,
        output_hidden_states: Optional[bool] = None,
        interpolate_pos_encoding: Optional[bool] = None,
        return_dict: Optional[bool] = None,
        cls_feat: bool = True,
        task_id: str = defaults.TASK_ID,
    ) -> Union[Tuple, BaseModelOutputWithPooling]:
        """Output has patch embeddings and the pooled output. We extract pooled CLS out by
        taking the second element.
        """
        out_to_filter = super().forward(
            pixel_values,
            head_mask,
            output_attentions,
            output_hidden_states,
            interpolate_pos_encoding,
            return_dict,
        )
        if isinstance(out_to_filter, BaseModelOutputWithPooling):
            return out_to_filter.pooler_output if cls_feat else out_to_filter.last_hidden_state
        return out_to_filter[0][:, 0] if cls_feat else out_to_filter[0]


class VisionTransformer(RenateBenchmarkingModule):
    """Vision Transformer base model.

    TODO: Fix citation
    Dosovitskiy, Alexey, et al. "An image is worth 16x16 words: Transformers for image recognition
    at scale."
    arXiv preprint arXiv:2010.11929 (2020).

    Args:
        pretrained_model_name_or_path: A string that denotes which pretrained model from the HF hub
            to use. If provided, it overrides other arguments about architecture.
        image_size: Size of the input image.
        patch_size: Size of the patches.
        num_layers: Number of Encoder layers.
        num_heads: Number of Attention heads.
        hidden_dim: Size of the Encoder's hidden state.
        mlp_dim: Size of the intermediate Multi-layer Perceptron in the Encoder.
        dropout: Dropout probability.
        attention_dropout: Dropout probability for the attention in the Multi-head Attention layer.
        num_outputs: Size of the output.
        prediction_strategy: Continual learning strategies may alter the prediction at train or test
            time.
        add_icarl_class_means: If ``True``, additional parameters used only by the
            ``ICaRLModelUpdater`` are added. Only required when using that updater.
    """

    def __init__(
        self,
        pretrained_model_name_or_path: Optional[str] = None,
        image_size: int = 32,
        patch_size: int = 4,
        num_layers: int = 12,
        num_heads: int = 12,
        hidden_dim: int = 768,
        mlp_dim: int = 3072,
        dropout: float = 0.1,
        attention_dropout: float = 0.1,
        num_outputs: int = 10,
        prediction_strategy: Optional[PredictionStrategy] = None,
        add_icarl_class_means: bool = True,
    ) -> None:
        if pretrained_model_name_or_path:
            model = FeatureExtractorViTModel.from_pretrained(
                pretrained_model_name_or_path=pretrained_model_name_or_path,
                return_dict=False,
                add_pooling_layer=False,
            )

            constructor_args = dict()
        else:
            model_config = ViTConfig(
                hidden_size=hidden_dim,
                num_hidden_layers=num_layers,
                num_attention_heads=num_heads,
                intermediate_size=mlp_dim,
                hidden_act="gelu",
                hidden_dropout_prob=dropout,
                attention_probs_dropout_prob=attention_dropout,
                layer_norm_eps=1e-6,
                image_size=image_size,
                patch_size=patch_size,
                num_channels=3,
                qkv_bias=True,
                return_dict=False,
            )

            model = FeatureExtractorViTModel(config=model_config, add_pooling_layer=False)
            constructor_args = {
                "image_size": image_size,
                "patch_size": patch_size,
                "num_layers": num_layers,
                "num_heads": num_heads,
                "hidden_dim": hidden_dim,
                "mlp_dim": mlp_dim,
                "dropout": dropout,
                "attention_dropout": attention_dropout,
            }

        super().__init__(
            embedding_size=model.config.hidden_size,
            num_outputs=num_outputs,
            constructor_arguments=constructor_args,
            prediction_strategy=prediction_strategy,
            add_icarl_class_means=add_icarl_class_means,
        )
        self._backbone = model

    def get_features(self, *args, **kwargs):
        # This is need as a shortcut to not call the base class's forward and directly call the
        # backbone's forward. Used only in L2P.
        return self._backbone(*args, **kwargs)


class VisionTransformerCIFAR(VisionTransformer):
    def __init__(self, **kwargs: Any) -> None:
        super().__init__(
            image_size=32,
            patch_size=4,
            num_layers=3,
            num_heads=3,
            hidden_dim=192,
            mlp_dim=768,
            **kwargs,
        )


class VisionTransformerB16(VisionTransformer):
    def __init__(self, **kwargs: Any) -> None:
        super().__init__(
            pretrained_model_name_or_path="google/vit-base-patch16-224",
            **kwargs,
        )


class VisionTransformerB32(VisionTransformer):
    def __init__(self, **kwargs: Any) -> None:
        super().__init__(
            pretrained_model_name_or_path="google/vit-base-patch32-224-in21k",
            **kwargs,
        )


class VisionTransformerL16(VisionTransformer):
    def __init__(self, **kwargs: Any) -> None:
        super().__init__(
            pretrained_model_name_or_path="google/vit-large-patch16-224-in21k",
            **kwargs,
        )


class VisionTransformerL32(VisionTransformer):
    def __init__(self, **kwargs: Any) -> None:
        super().__init__(
            pretrained_model_name_or_path="google/vit-large-patch32-224-in21k",
            **kwargs,
        )


class VisionTransformerH14(VisionTransformer):
    def __init__(self, **kwargs: Any) -> None:
        super().__init__(
            pretrained_model_name_or_path="google/vit-huge-patch14-224-in21k",
            **kwargs,
        )



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
import pytest
import torch

from renate.shift.detector import ShiftDetectorWithFeatureExtractor


@pytest.mark.parametrize(
    "dataset",
    [
        torch.utils.data.TensorDataset(torch.randn(20, 2)),
        torch.utils.data.TensorDataset(torch.randn(20, 2), torch.randint(0, 10, size=(20,))),
    ],
)
@pytest.mark.parametrize("feature_extractor", [torch.nn.Linear(2, 4)])
def test_extract_features(dataset, feature_extractor):
    detector = ShiftDetectorWithFeatureExtractor(feature_extractor=feature_extractor)
    features = detector.extract_features(dataset)
    assert features.size() == (20, 4)



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
import logging
from functools import partial
from typing import Any, Callable, Dict, List, Optional, Tuple

import torch
import torch.nn as nn
import torchmetrics
from pytorch_lightning.loggers.logger import Logger
from pytorch_lightning.utilities.types import STEP_OUTPUT
from torch.nn import Parameter
from torch.optim import Optimizer
from torch.optim.lr_scheduler import _LRScheduler

from renate import defaults
from renate.benchmark.models.l2p import LearningToPromptTransformer
from renate.models import RenateModule
from renate.types import NestedTensors
from renate.updaters.experimental.offline_er import OfflineExperienceReplayLearner
from renate.updaters.learner import Learner
from renate.updaters.model_updater import SingleTrainingLoopUpdater
from renate.utils.misc import maybe_populate_mask_and_ignore_logits

logger = logging.getLogger(__name__)


class LearningToPromptLearner(Learner):
    """Learner for learning to prompt

    This is identical to the base learner with an addition of loss term.
    TODO: Make this loss a component.

    Args:
        prompt_sim_loss_weight: Loss weight for the prompt key - image representation similarity
    """

    def __init__(
        self,
        prompt_sim_loss_weight: float = defaults.PROMPT_SIM_LOSS_WEIGHT,
        **kwargs,
    ) -> None:
        assert isinstance(
            kwargs["model"], LearningToPromptTransformer
        ), f"{self.__class__.__name__} can only train a LearningToPromptTransformer model"
        f"but got {type(kwargs['model'])}"
        super().__init__(
            **kwargs,
        )
        self.prompt_sim_loss_weight = prompt_sim_loss_weight
        self._loss_collections["train_losses"].update({"key_sim": torchmetrics.MeanMetric()})

    def training_step(
        self, batch: Tuple[NestedTensors, torch.Tensor], batch_idx: int
    ) -> STEP_OUTPUT:
        loss_dict = super().training_step(batch, batch_idx=batch_idx)
        key_similarity = -1 * self.prompt_sim_loss_weight * self._model.similarity_score
        loss_dict["loss"] += key_similarity
        self._loss_collections["train_losses"]["key_sim"](-key_similarity)
        return loss_dict


class LearningToPromptReplayLearner(OfflineExperienceReplayLearner):
    """L2P with an off-line ER learner.

    The model will be trained on weighted mixture of losses computed on the new data and a replay
    buffer. In contrast to the online version, the buffer will only be updated after training has
    terminated.

    Args:
        prompt_sim_loss_weight: Loss weight for the prompt key - image representation similarity
    """

    def __init__(
        self,
        prompt_sim_loss_weight: float = defaults.PROMPT_SIM_LOSS_WEIGHT,
        **kwargs,
    ) -> None:
        assert isinstance(
            kwargs["model"], LearningToPromptTransformer
        ), f"{self.__class__.__name__}  can only train a LearningToPromptTransformer model"
        f"but got {type(kwargs['model'])}"

        super().__init__(**kwargs)
        self.prompt_sim_loss_weight = prompt_sim_loss_weight
        self._loss_collections["train_losses"].update({"key_sim_loss": torchmetrics.MeanMetric()})

    def training_step(
        self, batch: Tuple[NestedTensors, torch.Tensor], batch_idx: int
    ) -> STEP_OUTPUT:
        """PyTorch Lightning function to return the training loss."""
        # The reason for rewriting is to ensure two independent forward props of inputs and memory
        # samples. LearningToPromptTransformer uses per_batch_prompt which uses a single prompt
        # repeated across the batch. Hence, the separate processing of memory and input samples.
        if self._loss_weight_new_data is None:
            alpha = self._num_points_current_task / (
                self._num_points_current_task + self._num_points_previous_tasks
            )
        else:
            alpha = self._loss_weight_new_data
        alpha = torch.tensor(alpha, device=batch["current_task"][0].device)
        inputs, targets = batch["current_task"]
        outputs = self(inputs)

        outputs, self._class_mask = maybe_populate_mask_and_ignore_logits(
            self._mask_unused_classes,
            self._class_mask,
            self._classes_in_current_task,
            outputs,
        )

        if "memory" in batch:
            (inputs_mem, targets_mem), _ = batch["memory"]
            outputs_mem = self(inputs_mem)

            outputs_mem, self._class_mask = maybe_populate_mask_and_ignore_logits(
                self._mask_unused_classes,
                self._class_mask,
                self._classes_in_current_task,
                outputs_mem,
            )

        loss_current = self._loss_fn(outputs, targets).mean()
        if "memory" in batch:
            loss_memory = self._loss_fn(outputs_mem, targets_mem).mean()
            self._loss_collections["train_losses"]["base_loss"](loss_current)
            self._loss_collections["train_losses"]["memory_loss"](loss_memory)
            loss = alpha * loss_current + (1.0 - alpha) * loss_memory
        else:
            loss = loss_current.mean()
            self._loss_collections["train_losses"]["base_loss"](loss)
        self._update_metrics(outputs, targets, "train")

        key_similarity = -1 * self.prompt_sim_loss_weight * self._model.similarity_score
        loss += key_similarity
        self._loss_collections["train_losses"]["key_sim"](-key_similarity)
        return {"loss": loss}


class LearningToPromptModelUpdater(SingleTrainingLoopUpdater):
    def __init__(
        self,
        model: RenateModule,
        loss_fn: torch.nn.Module,
        optimizer: Callable[[List[nn.Parameter]], Optimizer],
        batch_size: int = defaults.BATCH_SIZE,
        seed: int = defaults.SEED,
        learner_kwargs: Optional[Dict[str, Any]] = None,
        input_state_folder: Optional[str] = None,
        output_state_folder: Optional[str] = None,
        max_epochs: int = defaults.MAX_EPOCHS,
        learning_rate_scheduler: Optional[Optional[Callable[[Optimizer], _LRScheduler]]] = None,
        learning_rate_scheduler_interval: defaults.SUPPORTED_LR_SCHEDULER_INTERVAL_TYPE = defaults.LR_SCHEDULER_INTERVAL,  # noqa: E501
        prompt_sim_loss_weight: float = defaults.PROMPT_SIM_LOSS_WEIGHT,
        train_transform: Optional[Callable] = None,
        train_target_transform: Optional[Callable] = None,
        test_transform: Optional[Callable] = None,
        test_target_transform: Optional[Callable] = None,
        buffer_transform: Optional[Callable] = None,
        buffer_target_transform: Optional[Callable] = None,
        metric: Optional[str] = None,
        mode: defaults.SUPPORTED_TUNING_MODE_TYPE = "min",
        logged_metrics: Optional[Dict[str, torchmetrics.Metric]] = None,
        early_stopping_enabled: bool = False,
        logger: Logger = defaults.LOGGER(**defaults.LOGGER_KWARGS),
        accelerator: defaults.SUPPORTED_ACCELERATORS_TYPE = defaults.ACCELERATOR,
        devices: Optional[int] = None,
        strategy: Optional[str] = defaults.DISTRIBUTED_STRATEGY,
        precision: str = defaults.PRECISION,
        deterministic_trainer: bool = defaults.DETERMINISTIC_TRAINER,
        gradient_clip_val: Optional[float] = defaults.GRADIENT_CLIP_VAL,
        gradient_clip_algorithm: Optional[str] = defaults.GRADIENT_CLIP_ALGORITHM,
        mask_unused_classes: bool = defaults.MASK_UNUSED_CLASSES,
    ):
        learner_kwargs = {
            "batch_size": batch_size,
            "seed": seed,
            "loss_fn": loss_fn,
            "prompt_sim_loss_weight": prompt_sim_loss_weight,
        }
        super().__init__(
            model=model,
            loss_fn=loss_fn,
            optimizer=optimizer,
            learner_class=LearningToPromptLearner,
            learner_kwargs=learner_kwargs,
            input_state_folder=input_state_folder,
            output_state_folder=output_state_folder,
            max_epochs=max_epochs,
            learning_rate_scheduler=learning_rate_scheduler,
            learning_rate_scheduler_interval=learning_rate_scheduler_interval,
            train_transform=train_transform,
            train_target_transform=train_target_transform,
            test_transform=test_transform,
            test_target_transform=test_target_transform,
            buffer_transform=buffer_transform,
            buffer_target_transform=buffer_target_transform,
            metric=metric,
            mode=mode,
            logged_metrics=logged_metrics,
            early_stopping_enabled=early_stopping_enabled,
            logger=logger,
            accelerator=accelerator,
            devices=devices,
            strategy=strategy,
            precision=precision,
            deterministic_trainer=deterministic_trainer,
            gradient_clip_algorithm=gradient_clip_algorithm,
            gradient_clip_val=gradient_clip_val,
            mask_unused_classes=mask_unused_classes,
        )


class LearningToPromptReplayModelUpdater(SingleTrainingLoopUpdater):
    def __init__(
        self,
        model: RenateModule,
        loss_fn: torch.nn.Module,
        optimizer: Callable[[List[Parameter]], Optimizer],
        memory_size: int,
        batch_memory_frac: float = defaults.BATCH_MEMORY_FRAC,
        loss_weight_new_data: Optional[float] = None,
        learning_rate_scheduler: Optional[partial] = None,
        learning_rate_scheduler_interval: defaults.SUPPORTED_LR_SCHEDULER_INTERVAL_TYPE = defaults.LR_SCHEDULER_INTERVAL,  # noqa: E501
        prompt_sim_loss_weight: float = defaults.PROMPT_SIM_LOSS_WEIGHT,
        batch_size: int = defaults.BATCH_SIZE,
        input_state_folder: Optional[str] = None,
        output_state_folder: Optional[str] = None,
        max_epochs: int = defaults.MAX_EPOCHS,
        train_transform: Optional[Callable] = None,
        train_target_transform: Optional[Callable] = None,
        test_transform: Optional[Callable] = None,
        test_target_transform: Optional[Callable] = None,
        buffer_transform: Optional[Callable] = None,
        buffer_target_transform: Optional[Callable] = None,
        metric: Optional[str] = None,
        mode: defaults.SUPPORTED_TUNING_MODE_TYPE = "min",
        logged_metrics: Optional[Dict[str, torchmetrics.Metric]] = None,
        early_stopping_enabled: bool = False,
        logger: Logger = defaults.LOGGER(**defaults.LOGGER_KWARGS),
        accelerator: defaults.SUPPORTED_ACCELERATORS_TYPE = defaults.ACCELERATOR,
        devices: Optional[int] = None,
        strategy: str = defaults.DISTRIBUTED_STRATEGY,
        precision: str = defaults.PRECISION,
        seed: int = defaults.SEED,
        deterministic_trainer: bool = defaults.DETERMINISTIC_TRAINER,
        gradient_clip_val: Optional[float] = defaults.GRADIENT_CLIP_VAL,
        gradient_clip_algorithm: Optional[str] = defaults.GRADIENT_CLIP_ALGORITHM,
        mask_unused_classes: bool = defaults.MASK_UNUSED_CLASSES,
    ):
        learner_kwargs = {
            "memory_size": memory_size,
            "batch_memory_frac": batch_memory_frac,
            "loss_weight_new_data": loss_weight_new_data,
            "batch_size": batch_size,
            "seed": seed,
            "prompt_sim_loss_weight": prompt_sim_loss_weight,
        }
        super().__init__(
            model,
            loss_fn=loss_fn,
            optimizer=optimizer,
            learner_class=LearningToPromptReplayLearner,
            learner_kwargs=learner_kwargs,
            input_state_folder=input_state_folder,
            output_state_folder=output_state_folder,
            max_epochs=max_epochs,
            learning_rate_scheduler=learning_rate_scheduler,
            learning_rate_scheduler_interval=learning_rate_scheduler_interval,
            train_transform=train_transform,
            train_target_transform=train_target_transform,
            test_transform=test_transform,
            test_target_transform=test_target_transform,
            buffer_transform=buffer_transform,
            buffer_target_transform=buffer_target_transform,
            metric=metric,
            mode=mode,
            logged_metrics=logged_metrics,
            early_stopping_enabled=early_stopping_enabled,
            logger=logger,
            accelerator=accelerator,
            devices=devices,
            strategy=strategy,
            precision=precision,
            deterministic_trainer=deterministic_trainer,
            gradient_clip_algorithm=gradient_clip_algorithm,
            gradient_clip_val=gradient_clip_val,
            mask_unused_classes=mask_unused_classes,
        )



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
from typing import Optional

import torch


class Kernel:
    """Base class for kernel functions."""

    def __init__(self):
        pass

    def _check_inputs(self, X0: torch.Tensor, X1: torch.Tensor):
        assert X0.dim() == X1.dim() == 2
        assert X0.size(1) == X1.size(1)
        assert X0.dtype is X1.dtype


class RBFKernel(Kernel):
    """A radial basis function kernel.

    This kernel has one hyperparameter, a scalar lengthscale. If this is set to `None` (default),
    the lengthscale will be set adaptively, at _each_ call to the kernel, via the median heuristic.

    Args:
        lengthscale: The kernel lengthscale. If `None` (default), this is set automatically via the
            median heuristic. Note: In this case, the lengthscale will be reset at each call to the
            kernel.
    """

    def __init__(self, lengthscale: Optional[float] = None):
        super().__init__()
        self._lengthscale = lengthscale

    @torch.no_grad()
    def __call__(self, X0: torch.Tensor, X1: torch.Tensor):
        self._check_inputs(X0, X1)
        dists = torch.cdist(X0, X1)
        lengthscale = self._lengthscale or torch.median(dists)
        return torch.exp(-0.5 * dists**2 / lengthscale**2)



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
from renate.training import run_training_job


config_space = {
    "optimizer": "SGD",
    "momentum": 0.0,
    "weight_decay": 0.0,
    "learning_rate": 0.1,
    "alpha": 0.5,
    "batch_size": 64,
    "batch_memory_frac": 0.5,
    "memory_size": 500,
    "loss_normalization": 0,
    "loss_weight": 0.5,
    "early_stopping": True,
}

if __name__ == "__main__":
    # we run the first training job on the MNIST classes [0-4]
    run_training_job(
        config_space=config_space,
        mode="max",
        metric="val_accuracy",
        updater="ER",
        max_epochs=50,
        chunk_id=0,  # this selects the first chunk of the dataset
        config_file="renate_config.py",
        # this is where the model will be stored
        output_state_url="./state_dump_first_model/",
        # the training job will run on the local machine
        backend="local",
    )

    # retrieve the model from `./state_dump_first_model/` if you want
    # do not delete the model, we are going to use it below

    run_training_job(
        config_space=config_space,
        mode="max",
        metric="val_accuracy",
        updater="ER",
        max_epochs=50,
        chunk_id=1,  # this time we use the second chunk of the dataset
        config_file="renate_config.py",
        # the output of the first training job is loaded
        input_state_url="./state_dump_first_model/",
        # the new model will be stored in this folder
        output_state_url="./state_dump_second_model/",
        backend="local",
    )



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
import logging
import math
from typing import Any, Dict, List, Optional, Union

import torch
import torch.nn as nn

from renate.models.layers.shared_linear import SharedMultipleLinear
from renate.models.prediction_strategies import PredictionStrategy
from renate.models.task_identification_strategies import TaskPrototypes

from .l2p import PromptedTransformer
from .base import RenateBenchmarkingModule

logger = logging.getLogger(__name__)


class PromptPool(nn.Module):
    """Implements a pool of prompts to be used in for S-Prompts.

    Args:
        prompt_size: Equivalent to number of input tokens used per update . Defaults to 10.
        embedding_size: Hidden size of the transformer used.. Defaults to 768.
        current_update_id: Current update it. Used to init number of prompts. Defaults to 0.
    """

    def __init__(
        self, prompt_size: int = 10, embedding_size: int = 768, current_update_id: int = 0
    ) -> None:
        super().__init__()
        self._M = prompt_size
        self._embedding_size = embedding_size
        self._curr_task = current_update_id

        self._pool = nn.ParameterDict()
        for id in range(self._curr_task):
            # This always needs to be intialized as the torch's state dict only restores for an
            # existing Parameter.
            self._pool[f"{id}"] = nn.Parameter(
                torch.nn.init.kaiming_uniform_(
                    torch.empty((self._M, self._embedding_size)), a=math.sqrt(5)
                )
            )

        self._pool.requires_grad_(True)

    def forward(self, id: int) -> torch.nn.Parameter:
        return self._pool[f"{id}"]

    def get_params(self, id: int) -> List[torch.nn.Parameter]:
        return [self._pool[f"{id}"]]

    def increment_task(self) -> None:
        self._pool[f"{len(self._pool)}"] = nn.Parameter(
            torch.empty((self._M, self._embedding_size)).uniform_(-1, 1)
        )
        self._pool.requires_grad_(True)


class SPromptTransformer(RenateBenchmarkingModule):
    """Implements Transformer Model for S-Prompts as described in  Wang, Yabin, et.al ."S-prompts
       learning with pre-trained transformers: An occam’s razor for domain incremental learning."
       Advances in Neural Information Processing Systems 35 (2022): 5682-5695.

    Args:
        pretrained_model_name_or_path: A string that denotes which pretrained model from the HF hub
            to use.
        image_size: Image size. Used if `pretrained_model_name_or_path` is not set .
        patch_size: Patch size to be extracted. Used if `pretrained_model_name_or_path` is not set .
        num_layers: Num of transformer layers. Used only if `pretrained_model_name_or_path` is not
            set .
        num_heads: Num heads in MHSA. Used only if `pretrained_model_name_or_path` is not set .
        hidden_dim: Hidden dimension of transformers. Used only if `pretrained_model_name_or_path`
            is not set .
        mlp_dim: _description_. Used only if `pretrained_model_name_or_path` is not set .
        dropout: _description_. Used only if `pretrained_model_name_or_path` is not set .
        attention_dropout: _description_. Used only if `pretrained_model_name_or_path` is not set .
        num_outputs: Number of output classes of the output. Defaults to 10.
        prediction_strategy: Continual learning strategies may alter the prediction at train or test
            time. Defaults to None.
        add_icarl_class_means: If ``True``, additional parameters used only by the
            ``ICaRLModelUpdater`` are added. Only required when using that updater.
        prompt_size: Equivalent to number of input tokens used per update . Defaults to 10.
        task_id: Internal variable used to increment update id. Shouldn't be set by user.
            Defaults to 0.
        clusters_per_task: Number clusters in k-means used for task identification. Defaults to 5.
        per_task_classifier: Flag to share or use a common classifier head for all tasks.
            Defaults to False.
    """

    def __init__(
        self,
        pretrained_model_name_or_path="google/vit-base-patch16-224",
        image_size: int = 32,
        patch_size: int = 4,
        num_layers: int = 12,
        num_heads: int = 12,
        hidden_dim: int = 768,
        mlp_dim: int = 3072,
        dropout: float = 0.1,
        attention_dropout: float = 0.1,
        num_outputs: int = 10,
        prediction_strategy: Optional[PredictionStrategy] = None,
        add_icarl_class_means: bool = True,
        prompt_size: int = 10,
        task_id: int = 0,
        clusters_per_task: int = 5,
        per_task_classifier: bool = False,
    ):
        transformer = PromptedTransformer(
            pretrained_model_name_or_path=pretrained_model_name_or_path,
            image_size=image_size,
            patch_size=patch_size,
            num_layers=num_layers,
            num_heads=num_heads,
            hidden_dim=hidden_dim,
            mlp_dim=mlp_dim,
            dropout=dropout,
            attention_dropout=attention_dropout,
            num_outputs=num_outputs,
            prediction_strategy=prediction_strategy,
            add_icarl_class_means=add_icarl_class_means,
        )
        super().__init__(
            embedding_size=transformer.transformer._embedding_size,
            num_outputs=num_outputs,
            constructor_arguments=dict(
                **transformer.transformer._constructor_arguments,
                prompt_size=prompt_size,
                task_id=task_id,
                clusters_per_task=clusters_per_task,
                per_task_classifier=per_task_classifier,
            ),
            prediction_strategy=prediction_strategy,
            add_icarl_class_means=add_icarl_class_means,
        )
        self._M = prompt_size
        self._task_id = task_id
        self._per_task_classifier = per_task_classifier

        prompt_pool = PromptPool(
            prompt_size=self._M,
            embedding_size=self._embedding_size,
            current_update_id=self._task_id,
        )

        self._backbone = nn.ModuleDict({"transformer": transformer, "prompt_pool": prompt_pool})
        self._task_id_method = TaskPrototypes(
            task_id=task_id,
            clusters_per_task=clusters_per_task,
            embedding_size=self._embedding_size,
        )
        self._backbone["transformer"].requires_grad_(False)
        self._backbone["prompt_pool"].requires_grad_(True)

        self._backbone["classifier"] = SharedMultipleLinear(
            self._embedding_size,
            self._num_outputs,
            share_parameters=not self._per_task_classifier,
            num_updates=self._task_id + 1,
        )

        self._tasks_params = nn.ModuleDict(
            {k: nn.Identity() for k, _ in self._tasks_params.items()}
        )

        self._backbone.forward = self._forward_for_monkey_patching
        self.task_ids = None

    def increment_task(self) -> None:
        # This cannot be a part of add_task_params as the super.__init__ function calls
        # add_task_params, and thus we would be trying parameters to the non-existent
        # self.s_prompts
        self._backbone["prompt_pool"].increment_task()

    def _forward_for_monkey_patching(
        self, x: Union[torch.Tensor, Dict[str, Any]], task_id: str = None
    ) -> torch.Tensor:
        prompt = None
        task_ids = None
        if self.training:
            prompt = self._backbone["prompt_pool"](self._task_id)
        else:
            task_ids = self._task_id_method.infer_task(self._backbone["transformer"](x))
            if task_ids is not None:
                prompt = torch.stack([self._backbone["prompt_pool"](i) for i in task_ids])
                self.task_ids = task_ids.detach().cpu().numpy()

        features = self._backbone["transformer"](x, prompt)

        # additional logic for separate classifiers
        # a. This forward returns logits directly, and the RenateBenchmarkingModule's _task_params
        #    now are identities. Thus, the overall operation is still the network forward pass.
        # b. Additional handling of params is not needed as backbone's params will return all the
        #    necessary elements.

        if self.training:
            logits = self._backbone["classifier"][f"{self._task_id}"](features)
        elif task_ids is not None:
            logits = torch.cat(
                [
                    self._backbone["classifier"][f"{t}"](feat.unsqueeze(0))
                    for t, feat in zip(task_ids, features)
                ]
            )
        else:
            logits = self._backbone["classifier"]["0"](features)

        return logits

    def update_task_identifier(self, features: torch.Tensor, labels: torch.Tensor) -> None:
        self._task_id_method.update_task_prototypes(features, labels)

    def set_extra_state(self, state: Any, decode=True):
        super().set_extra_state(state, decode)
        # once this is set (after loading. increase that by one.)
        self._constructor_arguments["task_id"] = self._task_id + 1

    def features(self, x: torch.Tensor) -> torch.Tensor:
        return self._backbone["transformer"](x)



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
from pathlib import Path

import pandas as pd
import pytest

from renate.benchmark.experimentation import (
    cumulative_metrics_summary,
    execute_experiment_job,
    individual_metrics_summary,
)
from renate.evaluation.metrics.classification import average_accuracy


@pytest.fixture
def experiment_job_kwargs():
    return {
        "backend": "local",
        "config_file": str(
            Path(__file__).parent.parent / "renate_config_files" / "config_scenario.py"
        ),
        "config_space": {"updater": "ER", "max_epochs": 5},
        "mode": "max",
        "metric": "val_accuracy",
        "num_updates": 2,
        "max_time": 30,
        "seed": 0,
        "accelerator": "cpu",
        "devices": 1,
    }


@pytest.mark.parametrize("save_state", (True, False))
def test_execute_experiment_job(tmpdir, experiment_job_kwargs, save_state):
    """Only checking if things run, not testing anything besides that."""
    expected_columns = [
        "Task ID",
        "Average Accuracy",
        "Micro Average Accuracy",
        "Forgetting",
        "Forward Transfer",
        "Backward Transfer",
    ]
    expected_num_updates = experiment_job_kwargs["num_updates"]
    experiment_job_kwargs["save_state"] = save_state
    execute_experiment_job(experiment_outputs_url=tmpdir, **experiment_job_kwargs)
    results_df = pd.read_csv(str(Path(tmpdir) / "logs" / "metrics_summary.csv"))
    assert all(results_df.columns == expected_columns)
    if save_state:
        hpo_file = Path(tmpdir) / f"update_{expected_num_updates - 1}" / "hpo.csv"
        for update_id in range(expected_num_updates):
            assert (Path(tmpdir) / f"update_{update_id}" / "learner.ckpt").is_file()
            assert (Path(tmpdir) / f"update_{update_id}" / "model.ckpt").is_file()
    else:
        hpo_file = Path(tmpdir) / "hpo.csv"
    assert len(pd.read_csv(str(hpo_file))["update_id"].unique()) == expected_num_updates


@pytest.mark.parametrize(
    "update_dict,regex",
    [
        ({"backend": "UNKNOWN_BACKEND"}, r"Backend*"),
        ({"mode": "UNKNOWN_MODE"}, r"Mode*"),
        ({"num_updates": 3}, r"The dataset has*"),
    ],
    ids=["unknown backend", "unknown mode", "wrong number of updates"],
)
def test_execute_experiment_job_edge_cases(tmpdir, experiment_job_kwargs, update_dict, regex):
    """Check if basic input errors raise an exception."""
    experiment_job_kwargs.update(update_dict)
    with pytest.raises(AssertionError, match=regex):
        execute_experiment_job(experiment_outputs_url=tmpdir, **experiment_job_kwargs)


def test_cumulative_metrics_summary():
    results = {
        "accuracy": [[0.1, 0.2, 0.3], [0.4, 0.5, 0.6], [0.7, 0.8, 0.9]],
        "accuracy_init": [[0.1, 0.2, 0.3], [0.4, 0.5, 0.6], [0.7, 0.8, 0.9]],
    }
    metrics = [("Average Accuracy", average_accuracy)]
    df = cumulative_metrics_summary(
        results=results,
        cumulative_metrics=metrics,
        num_tasks=3,
        num_instances=[10, 20, 30],
    )
    assert list(df.columns) == ["Task ID", "Average Accuracy"]
    assert pytest.approx(list(df["Average Accuracy"])) == [0.1, 0.45, 0.8]
    assert df.shape == (3, 2)


def test_individual_metrics_summary():
    results = {
        "accuracy": [[0.1, 0.2, 0.3], [0.4, 0.5, 0.6], [0.7, 0.8, 0.9]],
        "accuracy_init": [[0.1, 0.2, 0.3], [0.4, 0.5, 0.6], [0.7, 0.8, 0.9]],
    }
    df = individual_metrics_summary(results=results, current_task=2, num_tasks=3)
    assert list(df.columns) == [("Task ID", "")] + [("accuracy", f"Task {i}") for i in range(1, 4)]
    assert list(df.iloc[0]) == [1.0, 0.1, 0.2, 0.3]
    assert list(df.iloc[1]) == [2.0, 0.4, 0.5, 0.6]
    assert df.shape == (2, 4)



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
from abc import ABC
from typing import Any, List, Optional

import torch
from torch import nn

from renate import defaults
from renate.models import RenateModule
from renate.models.prediction_strategies import ICaRLClassificationStrategy, PredictionStrategy
from renate.utils.deepspeed import convert_to_tensor, recover_object_from_tensor


# TODO: merge unit tests for the submodules
class RenateBenchmarkingModule(RenateModule, ABC):
    """Base class for all models provided by Renate.

    This class ensures that each models works with all ModelUpdaters when using the benchmarking
    feature of Renate. New models can extend this class or alternatively extend the RenateModule
    and make sure they are compatible with the considered ModelUpdater.

    Args:
        embedding_size: Representation size of the model after the backbone.
        num_outputs: The number of outputs of the model.
        constructor_arguments: Arguments needed to instantiate the model.
        prediction_strategy: By default a forward pass through the model. Some ModelUpdater must
            be combined with specific prediction strategies to work as intended.
        add_icarl_class_means: Specific parameters for iCaRL. Can be set to ``False`` if any other
            ModelUpdater is used.
    """

    def __init__(
        self,
        embedding_size: int,
        num_outputs: int,
        constructor_arguments: dict,
        prediction_strategy: Optional[PredictionStrategy] = None,
        add_icarl_class_means: bool = True,
    ):
        constructor_arguments["num_outputs"] = num_outputs
        constructor_arguments["add_icarl_class_means"] = add_icarl_class_means
        super().__init__(
            constructor_arguments=constructor_arguments,
        )
        self._embedding_size = embedding_size
        self._num_outputs = num_outputs
        self._prediction_strategy = prediction_strategy
        self._tasks_params: torch.nn.ModuleDict = torch.nn.ModuleDict()
        self.add_task_params(defaults.TASK_ID)
        if add_icarl_class_means:
            self.class_means = torch.nn.Parameter(
                torch.zeros((embedding_size, num_outputs)), requires_grad=False
            )

    def forward(self, x: torch.Tensor, task_id: str = defaults.TASK_ID) -> torch.Tensor:
        x = self.get_backbone(task_id=task_id)(x)
        if isinstance(self._prediction_strategy, ICaRLClassificationStrategy):
            return self._prediction_strategy(x, self.training, class_means=self.class_means)
        else:
            assert (
                self._prediction_strategy is None
            ), f"Unknown prediction strategy of type {type(self._prediction_strategy)}."
        return self.get_predictor(task_id)(x)

    def get_backbone(self, task_id: str = defaults.TASK_ID) -> nn.Module:
        """Returns the model without the prediction head."""
        return self._backbone

    def get_predictor(self, task_id: str = defaults.TASK_ID) -> nn.Module:
        """Returns the model without the backbone."""
        return self._tasks_params[task_id]

    def _add_task_params(self, task_id: str = defaults.TASK_ID) -> None:
        """Adds new parameters associated to a specific task to the model."""
        self._tasks_params[task_id] = torch.nn.Linear(self._embedding_size, self._num_outputs)

    def get_params(self, task_id: str = defaults.TASK_ID) -> List[torch.nn.Parameter]:
        """Returns the list of parameters for the core model and a specific `task_id`."""
        return list(self.get_backbone(task_id=task_id).parameters()) + list(
            self.get_predictor(task_id=task_id).parameters()
        )

    def get_extra_state(self, encode=True) -> Any:
        """Get the constructor_arguments and task ids necessary to reconstruct the model.

        Encode converts the state into a torch tensor so that Deepspeed serialization works.
        We don't encode any of the super() calls, but encode only the final dict.
        """
        extra_state = super().get_extra_state(encode=False)
        extra_state["prediction_strategy"] = self._prediction_strategy
        return convert_to_tensor(extra_state) if encode else extra_state

    def set_extra_state(self, state: Any, decode=True):
        """Extract the content of the ``_extra_state`` and set the related values in the module.

        decode flag is to decode the tensor of pkl bytes."""
        super().set_extra_state(state, decode=decode)
        self._prediction_strategy = (
            recover_object_from_tensor(state)["prediction_strategy"]
            if decode
            else state["prediction_strategy"]
        )



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
from typing import Optional

import torch
import transformers

import renate.defaults as defaults
from renate.benchmark.datasets.nlp_datasets import HuggingFaceTextDataModule
from renate.data.data_module import RenateDataModule
from renate.models import RenateModule
from renate.models.renate_module import RenateWrapper


def model_fn(model_state_url: Optional[str] = None) -> RenateModule:
    """Returns a DistilBert classification model."""
    transformer_model = transformers.DistilBertForSequenceClassification.from_pretrained(
        "distilbert-base-uncased", num_labels=2, return_dict=False
    )
    model = RenateWrapper(transformer_model)
    if model_state_url is not None:
        state_dict = torch.load(model_state_url)
        model.load_state_dict(state_dict)
    return model


def loss_fn() -> torch.nn.Module:
    return torch.nn.CrossEntropyLoss(reduction="none")


def data_module_fn(data_path: str, chunk_id: int, seed: int = defaults.SEED) -> RenateDataModule:
    """Returns one of two movie review datasets depending on `chunk_id`."""
    tokenizer = transformers.DistilBertTokenizer.from_pretrained("distilbert-base-uncased")
    dataset_name = "imdb" if chunk_id else "rotten_tomatoes"
    data_module = HuggingFaceTextDataModule(
        data_path,
        dataset_name=dataset_name,
        tokenizer=tokenizer,
        val_size=0.2,
        seed=seed,
    )
    return data_module



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
from typing import Optional

import torch
from torch.utils.data import DataLoader, Dataset

from renate.utils.pytorch import move_tensors_to_device


class ShiftDetector:
    """Base class for distribution shift detectors.

    The main interface consists of two methods `fit` and `score`, which expect pytorch Dataset
    objects. One passes a reference dataset to the `fit` method. Then we can check query datasets
    for distribution shifts (relative to the reference dataset) using the `score` method. The
    `score` method returns a scalar shift score with the convention that high values indicate a
    distribution shift. For most methods, this score will be in [0, 1].

    Args:
        batch_size: Batch size used to iterate over datasets, e.g., for extracting features. This
            choice does not affect the result of the shift detector, but might affect run time.
        num_preprocessing_workers: Number of workers used in data loaders.
        device: Device to use for computations inside the detector.
    """

    def __init__(
        self,
        batch_size: int = 32,
        num_preprocessing_workers: int = 0,
        device: str = "cpu",
    ) -> None:
        self._batch_size = batch_size
        self._num_preprocessing_workers = num_preprocessing_workers
        self._device = device

    def fit(self, dataset: Dataset) -> None:
        """Fit the detector to a reference dataset."""
        raise NotImplementedError()

    def score(self, dataset: Dataset) -> float:
        """Compute distribution shift score for a query dataset."""
        raise NotImplementedError()

    def _make_data_loader(self, dataset: Dataset, shuffle: bool = False) -> DataLoader:
        """Return a data loader to iterate over a dataset.

        Args:
            dataset: The dataset.
            shuffle: Whether to shuffle or not.
        """
        return DataLoader(
            dataset,
            batch_size=self._batch_size,
            shuffle=shuffle,
            num_workers=self._num_preprocessing_workers,
        )


class ShiftDetectorWithFeatureExtractor(ShiftDetector):
    """Base class for detectors working on extracted features.

    These shift detectors extract some (lower-dimensional) features from the datasets, which are
    used as inputs to the shift detection methods. Subclasses have to overwrite `fit_with_features`
    and `score_with_features`.

    Args:
        feature_extractor: A pytorch model used as feature extractor.
        batch_size: Batch size used to iterate over datasets.
        num_preprocessing_workers: Number of workers used in data loaders.
        device: Device to use for computations inside the detector.
    """

    def __init__(
        self,
        feature_extractor: Optional[torch.nn.Module] = None,
        batch_size: int = 32,
        num_preprocessing_workers: int = 0,
        device: str = "cpu",
    ) -> None:
        super(ShiftDetectorWithFeatureExtractor, self).__init__(
            batch_size, num_preprocessing_workers, device
        )
        self._feature_extractor = feature_extractor or torch.nn.Identity()
        self._feature_extractor = self._feature_extractor.to(self._device)

    def fit(self, dataset: Dataset) -> None:
        """Fit the detector to a reference dataset."""
        X = self.extract_features(dataset)
        self._fit_with_features(X)

    def score(self, dataset: Dataset) -> float:
        """Compute distribution shift score for a query dataset."""
        X = self.extract_features(dataset)
        return self._score_with_features(X)

    @torch.no_grad()
    def extract_features(self, dataset: Dataset) -> torch.Tensor:
        """Extract features from a dataset."""
        dataloader = self._make_data_loader(dataset)
        Xs = []
        for batch in dataloader:
            X = move_tensors_to_device(batch[0], device=self._device)
            Xs.append(self._feature_extractor(X))
        X = torch.cat(Xs, dim=0).cpu()
        return X

    def _fit_with_features(self, X: torch.Tensor) -> None:
        raise NotImplementedError()

    def _score_with_features(self, X: torch.Tensor) -> float:
        raise NotImplementedError()
